import hub
import time
import uuid
import struct
import atexit
import threading

from typing import Tuple, Dict, Callable, Optional
from hub.util.exceptions import LockedException
from hub.util.keys import get_dataset_lock_key
from hub.util.path import get_path_from_storage
from hub.util.threading import terminate_thread
from hub.core.storage import StorageProvider


class Lock(object):
    """Locks a StorageProvider instance to avoid concurrent writes from multiple machines.

    Example:
        From machine 1:
        s3 = hub.core.storage.S3Provider(S3_URL)
        lock = hub.core.lock.Lock(s3)  # Works

        From machine 2:
        s3 = hub.core.storage.S3Provider(S3_URL)
        lock = hub.core.lock.Lock(s3)  # Raises LockedException

        The lock is updated every 2 mins by an internal thread. The lock is valid for 5 mins after the last update.

    Args:
        storage (StorageProvider): The storage provder to be locked.
        callback (Callable, optional): Called if the lock is lost after acquiring.

    Raises:
        LockedException: If the storage is already locked by a different machine.
    """

    def __init__(self, storage: StorageProvider, callback: Optional[Callable] = None):
        self.storage = storage
        self.callback = callback
        self.acquired = False
        self._thread_lock = threading.Lock()
        self._previous_update_timestamp = None
        self.acquire()
        atexit.register(self.release)

    def _get_lock_bytes(self) -> bytes:
        return uuid.getnode().to_bytes(6, "little") + struct.pack("d", time.time())

    def _parse_lock_bytes(self, byts) -> Tuple[int, float]:
        byts = memoryview(byts)
        nodeid = int.from_bytes(byts[:6], "little")
        timestamp = struct.unpack("d", byts[6:])[0]
        return nodeid, timestamp

    def _lock_loop(self):
        try:
            while True:
                try:
                    if (
                        self._previous_update_timestamp is not None
                        and time.time() - self._previous_update_timestamp
                        >= hub.constants.DATASET_LOCK_VALIDITY
                    ):
                        # Its been too long since last update, another machine might have locked the storage
                        lock_bytes = self.storage.get(get_dataset_lock_key())
                        if lock_bytes:
                            nodeid, timestamp = self._parse_lock_bytes(lock_bytes)
                            if nodeid != uuid.getnode():
                                if self.callback:
                                    self.callback()
                                self.acquired = False
                                return
                    self._previous_update_timestamp = time.time()
                    self.storage[get_dataset_lock_key()] = self._get_lock_bytes()
                except Exception:
                    pass
                time.sleep(hub.constants.DATASET_LOCK_UPDATE_INTERVAL)
        except Exception:  # Thread termination
            return

    def acquire(self):
        if self.acquired:
            return
        self.storage.check_readonly()
        lock_bytes = self.storage.get(get_dataset_lock_key())
        if lock_bytes is not None:
            nodeid = None
            try:
                nodeid, timestamp = self._parse_lock_bytes(lock_bytes)
            except Exception:  # parse error from corrupt lock file, ignore
                pass
            if nodeid:
                if nodeid == uuid.getnode():
                    # Lock left by this machine from a previous run, ignore
                    pass
                elif time.time() - timestamp < hub.constants.DATASET_LOCK_VALIDITY:
                    raise LockedException()

        self._thread = threading.Thread(target=self._lock_loop, daemon=True)
        self._thread.start()
        self.acquired = True

    def release(self):
        if not self.acquired:
            return
        with self._thread_lock:
            terminate_thread(self._thread)
            self._acquired = False
        try:
            del self.storage[get_dataset_lock_key()]
        except Exception:
            pass


_LOCKS: Dict[str, Lock] = {}


def lock(storage: StorageProvider, callback: Optional[Callable] = None):
    """Locks a StorageProvider instance to avoid concurrent writes from multiple machines.

    Args:
        storage (StorageProvider): The storage provder to be locked.
        callback (Callable, optional): Called if the lock is lost after acquiring.

    Raises:
        LockedException: If the storage is already locked by a different machine.
    """

    path = get_path_from_storage(storage)
    lock = _LOCKS.get(path)
    if lock:
        lock.acquire()
    else:
        lock = Lock(storage, callback=callback)
        _LOCKS[path] = lock


def unlock(storage: StorageProvider):
    """Unlocks a storage provider that was locked by this machine.

    Args:
        storage (StorageProvider): The storage provder to be locked.
    """
    path = get_path_from_storage(storage)
    lock = _LOCKS.get(path)
    if lock:
        lock.release()



from hub.util.auto import ingestion_summary
from hub.api.dataset import Dataset
from hub.tests.common import get_dummy_data_path
from hub.util.exceptions import (
    InvalidPathException,
    SamePathException,
    TensorAlreadyExistsError,
)
import pytest
import hub


def test_ingestion_simple(memory_ds: Dataset):
    path = get_dummy_data_path("tests_auto/image_classification")

    with pytest.raises(InvalidPathException):
        hub.ingest(
            src="tests_auto/invalid_path",
            dest=memory_ds.path,
            images_compression="auto",
            progress_bar=False,
            summary=False,
            overwrite=False,
        )

    with pytest.raises(SamePathException):
        hub.ingest(
            src=path,
            dest=path,
            images_compression="jpeg",
            progress_bar=False,
            summary=False,
            overwrite=False,
        )

    ds = hub.ingest(
        src=path,
        dest=memory_ds.path,
        images_compression="auto",
        progress_bar=False,
        summary=False,
        overwrite=False,
    )

    assert ds["images"].meta.sample_compression == "jpeg"
    assert list(ds.tensors.keys()) == ["images", "labels"]
    assert ds["images"].numpy().shape == (3, 200, 200, 3)
    assert ds["labels"].numpy().shape == (3, 1)
    assert ds["labels"].info.class_names == ("class0", "class1", "class2")


def test_image_classification_sets(memory_ds: Dataset):
    path = get_dummy_data_path("tests_auto/image_classification_with_sets")
    ds = hub.ingest(
        src=path,
        dest=memory_ds.path,
        images_compression="auto",
        progress_bar=False,
        summary=False,
        overwrite=False,
    )

    assert list(ds.tensors) == [
        "test/images",
        "test/labels",
        "train/images",
        "train/labels",
    ]

    assert ds["train/images"].meta.sample_compression == "jpeg"
    assert ds["test/images"].numpy().shape == (3, 200, 200, 3)
    assert ds["test/labels"].numpy().shape == (3, 1)
    assert ds["test/labels"].info.class_names == ("class0", "class1", "class2")

    assert ds["train/images"].numpy().shape == (3, 200, 200, 3)
    assert ds["train/labels"].numpy().shape == (3, 1)
    assert ds["train/labels"].info.class_names == ("class0", "class1", "class2")


def test_ingestion_exception(memory_ds: Dataset):
    path = get_dummy_data_path("tests_auto/image_classification_with_sets")
    with pytest.raises(InvalidPathException):
        hub.ingest(
            src="tests_auto/invalid_path",
            dest=memory_ds.path,
            images_compression="auto",
            progress_bar=False,
            summary=False,
            overwrite=False,
        )

    with pytest.raises(SamePathException):
        hub.ingest(
            src=path,
            dest=path,
            images_compression="auto",
            progress_bar=False,
            summary=False,
            overwrite=False,
        )


def test_overwrite(local_ds: Dataset):
    path = get_dummy_data_path("tests_auto/image_classification")

    hub.ingest(
        src=path,
        dest=local_ds.path,
        images_compression="auto",
        progress_bar=False,
        summary=False,
        overwrite=False,
    )

    with pytest.raises(TensorAlreadyExistsError):
        hub.ingest(
            src=path,
            dest=local_ds.path,
            images_compression="auto",
            progress_bar=False,
            summary=False,
            overwrite=False,
        )



import sys
import hub
import pytest
from io import StringIO
from contextlib import contextmanager
from click.testing import CliRunner
from hub.cli.auth import login, logout
from hub.util.agreement import get_all_local_agreements, update_local_agreements
from hub.util.exceptions import AgreementNotAcceptedError, NotLoggedInError


@contextmanager
def replace_stdin(target):
    orig = sys.stdin
    sys.stdin = target
    yield
    sys.stdin = orig


def dont_agree(path):
    """Load the hub cloud dataset at path and simulate disagreeing to the terms of access."""

    with pytest.raises(AgreementNotAcceptedError):
        # this text can be anything except expected
        with replace_stdin(StringIO("no, i don't agree!")):
            hub.load(path)


def agree(path):
    """Load the hub cloud dataset at path and simulate agreeing to the terms of access."""

    dataset_name = path.split("/")[-1]
    with replace_stdin(StringIO(dataset_name)):
        ds = hub.load(path)
    ds.labels[0].numpy()


def remove_agreement(username, path):
    """Removes the agreement for path from the locally stored info."""
    all_local_agreements = get_all_local_agreements()
    agreement_set = all_local_agreements.get(username) or set()
    agreement_set.discard(path)
    all_local_agreements[username] = agreement_set
    update_local_agreements(all_local_agreements)


def test_agreement_logged_out(hub_cloud_dev_credentials):
    runner = CliRunner()
    runner.invoke(logout)
    path = "hub://activeloop/imagenet-train"
    with pytest.raises(NotLoggedInError):
        agree(path)


def test_agreement_logged_in(hub_cloud_dev_credentials):
    runner = CliRunner()
    username, password = hub_cloud_dev_credentials
    runner.invoke(login, f"-u {username} -p {password}")
    path = "hub://activeloop/imagenet-train"
    agree(path)
    runner.invoke(logout)
    remove_agreement(username, path)


def test_not_agreement_logged_in(hub_cloud_dev_credentials):
    runner = CliRunner()
    username, password = hub_cloud_dev_credentials
    runner.invoke(login, f"-u {username} -p {password}")
    path = "hub://activeloop/imagenet-train"
    dont_agree(path)
    runner.invoke(logout)



import numpy as np
import hub
import pytest
from hub.util.json import JsonValidationError
from hub.tests.dataset_fixtures import enabled_non_gcs_datasets
from typing import Any, Optional, Union, List, Dict


def test_json_basic(memory_ds):
    ds = memory_ds
    ds.create_tensor("json", htype="json")
    items = [
        {"x": [1, 2, 3], "y": [4, [5, 6]]},
        {"x": [1, 2, 3], "y": [4, {"z": [0.1, 0.2, []]}]},
    ]
    with ds:
        for x in items:
            ds.json.append(x)
        ds.json.extend(items)
    assert ds.json.shape == (4, 1)
    for i in range(4):
        assert ds.json[i].data() == items[i % 2]


def test_json_with_numpy(memory_ds):
    ds = memory_ds
    ds.create_tensor("json", htype="json")
    items = [
        {"x": np.array([1, 2, 3], dtype=np.float32), "y": [4, [5, 6]]},
        {"x": np.array([1, 2, 3], dtype=np.uint8), "y": [4, {"z": [0.1, 0.2, []]}]},
    ]
    with ds:
        for x in items:
            ds.json.append(x)
        ds.json.extend(items)
    for i in range(4):
        assert ds.json[i].data()["y"] == items[i % 2]["y"]
        np.testing.assert_array_equal(ds.json[i].data()["x"], items[i % 2]["x"])


def test_json_with_hub_sample(memory_ds, compressed_image_paths):
    ds = memory_ds
    ds.create_tensor("json", htype="json")
    items = [
        {
            "x": [1, 2, 3],
            "y": [4, [5, 6]],
            "z": hub.read(compressed_image_paths["jpeg"][0]),
        },
        {
            "x": [1, 2, 3],
            "y": [4, {"z": [0.1, 0.2, []]}],
            "z": hub.read(compressed_image_paths["png"][0]),
        },
    ]
    with ds:
        for x in items:
            ds.json.append(x)
        ds.json.extend(items)
    assert ds.json.shape == (4, 1)
    for i in range(4):
        assert ds.json[i].data() == items[i % 2]


def test_json_list_basic(memory_ds):
    ds = memory_ds
    ds.create_tensor("list", htype="list")
    items = [
        [{"x": [1, 2, 3], "y": [4, [5, 6]]}, [[]], [None, 0.1]],
        [[], [[[]]], {"a": [0.1, 1, "a", []]}],
    ]
    with ds:
        for x in items:
            ds.list.append(x)
        ds.list.extend(items)
    assert ds.list.shape == (4, 3)
    for i in range(4):
        assert ds.list[i].data() == items[i % 2]
    for i, x in enumerate(ds.list.data()):
        assert x == items[i % 2]


def test_list_with_numpy(memory_ds):
    ds = memory_ds
    ds.create_tensor("list", htype="list")
    items = [
        [
            np.random.random((3, 4)),
            {"x": [1, 2, 3], "y": [4, [5, 6]]},
            [[]],
            [None, 0.1],
        ],
        [np.random.randint(0, 10, (4, 5)), [], [[[]]], {"a": [0.1, 1, "a", []]}],
    ]
    with ds:
        for x in items:
            ds.list.append(x)
        ds.list.extend(items)
    assert ds.list.shape == (4, 4)
    for i in range(4):
        actual, expected = ds.list[i].data(), items[i % 2]
        np.testing.assert_array_equal(actual[0], expected[0])
        assert actual[1:] == expected[1:]


def test_list_with_hub_sample(memory_ds, compressed_image_paths):
    ds = memory_ds
    ds.create_tensor("list", htype="list")
    items = [
        [
            {
                "x": [1, 2, 3],
                "y": [4, [5, 6, hub.read(compressed_image_paths["jpeg"][0])]],
            },
            [[hub.read(compressed_image_paths["jpeg"][1])]],
            [None, 0.1],
        ],
        [
            [],
            [[[hub.read(compressed_image_paths["png"][0])]]],
            {"a": [0.1, 1, "a", hub.read(compressed_image_paths["png"][0])]},
        ],
    ]
    with ds:
        for x in items:
            ds.list.append(x)
        ds.list.extend(items)
    assert ds.list.shape == (4, 3)
    for i in range(4):
        assert ds.list[i].data() == items[i % 2]


def test_json_with_schema(memory_ds):
    ds = memory_ds
    ds.create_tensor("json", htype="json", dtype=List[Dict[str, int]])
    ds.json.append([{"x": 1, "y": 2}])
    with pytest.raises(JsonValidationError):
        ds.json.append({"x": 1, "y": 2})
    with pytest.raises(JsonValidationError):
        ds.json.append([{"x": 1, "y": "2"}])

    assert ds.json.numpy()[0, 0] == [{"x": 1, "y": 2}]

    ds.create_tensor("json2", htype="json", dtype=Optional[List[Dict[str, int]]])
    items = [
        [{"x": 1, "y": 2}],
        None,
        [{"x": 2, "y": 3}],
        None,
    ]
    ds.json2.extend(items)
    for i in range(len(items)):
        assert ds.json2[i].data() == ds.json2.data()[i] == items[i]


@enabled_non_gcs_datasets
@pytest.mark.parametrize("compression", ["lz4", None])
def test_json_transform(ds, compression, scheduler="threaded"):
    ds.create_tensor("json", htype="json", sample_compression=compression)

    items = [
        {"x": [1, 2, 3], "y": [4, [5, 6]]},
        {"x": [1, 2, 3], "y": [4, {"z": [0.1, 0.2, []]}]},
        ["a", ["b", "c"], {"d": 1.0}],
        [1.0, 2.0, 3.0, 4.0],
        ["a", "b", "c", "d"],
        1,
        5.0,
        True,
        False,
        None,
    ] * 5

    @hub.compute
    def upload(stuff, ds):
        ds.json.append(stuff)
        return ds

    upload().eval(items, ds, num_workers=2, scheduler=scheduler)
    assert ds.json.data() == items


@enabled_non_gcs_datasets
@pytest.mark.parametrize("compression", ["lz4", None])
def test_list_transform(ds, compression, scheduler="threaded"):
    ds.create_tensor("list", htype="list")

    items = [
        ["a", ["b", "c"], {"d": 1.0}],
        [1.0, 2.0, 3.0, 4.0],
        ["a", "b", "c", "d"],
    ] * 5

    @hub.compute
    def upload(stuff, ds):
        ds.list.append(stuff)
        return ds

    upload().eval(items, ds, num_workers=2, scheduler=scheduler)
    assert ds.list.data() == items



from hub.util.exceptions import DatasetHandlerError
from click.testing import CliRunner
import pytest
import hub
import numpy as np


def test_new_dataset():
    with CliRunner().isolated_filesystem():
        ds = hub.dataset("test_new_dataset")
        with ds:
            ds.create_tensor("image")
            for i in range(10):
                ds.image.append(i * np.ones((100 * (i + 1), 100 * (i + 1))))

        for i in range(10):
            np.testing.assert_array_equal(
                ds.image[i].numpy(), i * np.ones((100 * (i + 1), 100 * (i + 1)))
            )


def test_dataset_empty_load():
    with CliRunner().isolated_filesystem():
        path = "test_dataset_load"

        ds = hub.empty(path)
        with ds:
            ds.create_tensor("image")
            for i in range(10):
                ds.image.append(i * np.ones((100 * (i + 1), 100 * (i + 1))))

        with pytest.raises(DatasetHandlerError):
            ds_empty = hub.empty(path)

        ds_loaded = hub.load(path)
        for i in range(10):
            np.testing.assert_array_equal(
                ds.image[i].numpy(), ds_loaded.image[i].numpy()
            )

        with pytest.raises(DatasetHandlerError):
            ds_random = hub.load("some_random_path")

        ds_overwrite_load = hub.dataset(path, overwrite=True)
        assert len(ds_overwrite_load) == 0
        assert len(ds_overwrite_load.tensors) == 0
        with ds_overwrite_load:
            ds_overwrite_load.create_tensor("image")
            for i in range(10):
                ds_overwrite_load.image.append(
                    i * np.ones((100 * (i + 1), 100 * (i + 1)))
                )

        with pytest.raises(DatasetHandlerError):
            ds_empty = hub.empty(path)

        ds_overwrite_empty = hub.dataset(path, overwrite=True)
        assert len(ds_overwrite_empty) == 0
        assert len(ds_overwrite_empty.tensors) == 0


def test_update_privacy(hub_cloud_ds):
    assert not hub_cloud_ds.public
    hub_cloud_ds.make_public()
    assert hub_cloud_ds.public
    hub_cloud_ds.make_private()
    assert not hub_cloud_ds.public
    with pytest.raises(hub.util.exceptions.AuthorizationException):
        hub.load(hub_cloud_ds.path)


def test_persistence_bug(local_ds_generator):
    for tensor_name in ["abc", "abcd/defg"]:
        ds = local_ds_generator()
        with ds:
            ds.create_tensor(tensor_name)
            ds[tensor_name].append(1)

        ds = local_ds_generator()
        with ds:
            ds[tensor_name].append(2)

        ds = local_ds_generator()
        np.testing.assert_array_equal(ds[tensor_name].numpy(), np.array([[1], [2]]))



from hub.constants import (
    HUB_CLOUD_OPT,
    ENV_HUB_DEV_USERNAME,
    ENV_HUB_DEV_PASSWORD,
    ENV_KAGGLE_USERNAME,
    ENV_KAGGLE_KEY,
    KAGGLE_OPT,
)
from hub.tests.common import is_opt_true
import os
import pytest
from hub.client.client import HubBackendClient
from hub.client.config import USE_LOCAL_HOST, USE_DEV_ENVIRONMENT

from warnings import warn


@pytest.fixture(scope="session")
def hub_cloud_dev_credentials(request):
    if not is_opt_true(request, HUB_CLOUD_OPT):
        pytest.skip()

    if not (USE_LOCAL_HOST or USE_DEV_ENVIRONMENT):
        warn(
            "Running hub cloud tests without setting USE_LOCAL_HOST or USE_DEV_ENVIRONMENT is not recommended."
        )

    username = os.getenv(ENV_HUB_DEV_USERNAME)
    password = os.getenv(ENV_HUB_DEV_PASSWORD)

    assert (
        username is not None
    ), f"Hub dev username was not found in the environment variable '{ENV_HUB_DEV_USERNAME}'. This is necessary for testing hub cloud datasets."
    assert (
        password is not None
    ), f"Hub dev password was not found in the environment variable '{ENV_HUB_DEV_PASSWORD}'. This is necessary for testing hub cloud datasets."

    return username, password


@pytest.fixture(scope="session")
def hub_cloud_dev_token(hub_cloud_dev_credentials):
    username, password = hub_cloud_dev_credentials

    client = HubBackendClient()
    token = client.request_auth_token(username, password)
    return token


@pytest.fixture(scope="session")
def hub_kaggle_credentials(request):
    if not is_opt_true(request, KAGGLE_OPT):
        pytest.skip()

    username = os.getenv(ENV_KAGGLE_USERNAME)
    key = os.getenv(ENV_KAGGLE_KEY)

    assert (
        key is not None
    ), f"Kaggle credentials were not found in environment variable. This is necessary for testing kaggle ingestion datasets."

    return username, key



from hub.core.sample import Sample  # type: ignore
from hub.util.exceptions import TensorInvalidSampleShapeError
import numpy as np


class TransformTensor:
    def __init__(self, base_tensor=None, slice_list=None) -> None:
        self.items = [] if base_tensor is None else base_tensor.items
        self.base_tensor = base_tensor or self
        self.slice_list = slice_list or []
        self.length = None

    def numpy(self) -> None:
        """Returns all the items stored in the slice of the tensor as numpy arrays. Even samples stored using hub.read are converted to numpy arrays in this."""
        value = self.numpy_compressed()
        if isinstance(value, list):
            for i in range(len(value)):
                if isinstance(value[i], Sample):
                    value[i] = value[i].array
        elif isinstance(value, Sample):
            value = value.array
        return value

    def numpy_compressed(self):
        """Returns all the items stored in the slice of the tensor. Samples stored using hub.read are not converted to numpy arrays in this."""
        value = self.items
        for slice_ in self.slice_list:
            value = value[slice_]
        self.length = len(value) if isinstance(value, list) else 1
        return value

    def __len__(self) -> int:
        if self.length is None:
            self.numpy_compressed()  # calculates and sets length
        return self.length  # type: ignore

    def __getitem__(self, index):
        if isinstance(index, list):
            new_slice_list = self.slice_list + index
        else:
            new_slice_list = self.slice_list + [index]
        return TransformTensor(base_tensor=self.base_tensor, slice_list=new_slice_list)

    def append(self, item):
        """Adds an item to the tensor."""
        if not isinstance(item, Sample):
            item = np.asarray(item)
            if self.items:
                expected_dims = self.items[-1].ndim
                dims = item.ndim
                if expected_dims != dims:
                    raise TensorInvalidSampleShapeError(item.shape, expected_dims)

        self.items.append(item)

    def extend(self, items):
        """Adds multiple items to the tensor."""
        for item in items:
            self.append(item)



from hub.core.meta.encode.base_encoder import Encoder, LAST_SEEN_INDEX_COLUMN
from hub.constants import ENCODING_DTYPE, UUID_SHIFT_AMOUNT
from hub.util.exceptions import ChunkIdEncoderError
import hub
from hub.core.storage.cachable import Cachable
import numpy as np
from uuid import uuid4
from hub.core.serialize import serialize_chunkids, deserialize_chunkids


CHUNK_ID_COLUMN = 0


class ChunkIdEncoder(Encoder, Cachable):
    def tobytes(self) -> memoryview:
        return serialize_chunkids(self.version, [self._encoded])

    @staticmethod
    def name_from_id(id: ENCODING_DTYPE) -> str:
        """Returns the hex of `id` with the "0x" prefix removed. This is the chunk's name and should be used to determine the chunk's key.
        Can convert back into `id` using `id_from_name`. You can get the `id` for a chunk using `__getitem__`."""

        return hex(id)[2:]

    @staticmethod
    def id_from_name(name: str):
        """Returns the 64-bit integer from the hex `name` generated by `name_from_id`."""

        return int("0x" + name, 16)

    def get_name_for_chunk(self, chunk_index: int) -> str:
        """Gets the name for the chunk at index `chunk_index`. If you need to get the name for a chunk from a sample index, instead
        use `__getitem__`, then `name_from_id`."""

        chunk_id = self._encoded[:, CHUNK_ID_COLUMN][chunk_index]
        return ChunkIdEncoder.name_from_id(chunk_id)

    @classmethod
    def frombuffer(cls, buffer: bytes):
        instance = cls()
        if not buffer:
            return instance
        version, ids = deserialize_chunkids(buffer)
        if ids.nbytes:
            instance._encoded = ids
        instance.version = version
        return instance

    @property
    def num_chunks(self) -> int:
        if self.num_samples == 0:
            return 0
        return len(self._encoded)

    def generate_chunk_id(self) -> ENCODING_DTYPE:
        """Generates a random 64bit chunk ID using uuid4. Also prepares this ID to have samples registered to it.
        This method should be called once per chunk created.

        Returns:
            ENCODING_DTYPE: The random chunk ID.
        """

        id = ENCODING_DTYPE(uuid4().int >> UUID_SHIFT_AMOUNT)

        if self.num_samples == 0:
            self._encoded = np.array([[id, -1]], dtype=ENCODING_DTYPE)

        else:
            last_index = self.num_samples - 1

            new_entry = np.array(
                [[id, last_index]],
                dtype=ENCODING_DTYPE,
            )
            self._encoded = np.concatenate([self._encoded, new_entry])

        return id

    def register_samples(self, num_samples: int):  # type: ignore
        """Registers samples to the chunk ID that was generated last with the `generate_chunk_id` method.
        This method should be called at least once per chunk created.

        Args:
            num_samples (int): The number of samples the last chunk ID should have added to it's registration.

        Raises:
            ValueError: `num_samples` should be non-negative.
            ChunkIdEncoderError: Must call `generate_chunk_id` before registering samples.
            ChunkIdEncoderError: `num_samples` can only be 0 if it is able to be a sample continuation accross chunks.
        """

        super().register_samples(None, num_samples)

    def translate_index_relative_to_chunks(self, global_sample_index: int) -> int:
        """Converts `global_sample_index` into a new index that is relative to the chunk the sample belongs to.

        Example:
            Given: 2 sampes in chunk 0, 2 samples in chunk 1, and 3 samples in chunk 2.
            >>> self.num_samples
            7
            >>> self.num_chunks
            3
            >>> self.translate_index_relative_to_chunks(0)
            0
            >>> self.translate_index_relative_to_chunks(1)
            1
            >>> self.translate_index_relative_to_chunks(2)
            0
            >>> self.translate_index_relative_to_chunks(3)
            1
            >>> self.translate_index_relative_to_chunks(6)
            2

        Args:
            global_sample_index (int): Index of the sample relative to the containing tensor.

        Returns:
            int: local index value between 0 and the amount of samples the chunk contains - 1.
        """

        _, chunk_index = self.__getitem__(global_sample_index, return_row_index=True)  # type: ignore

        if chunk_index == 0:
            return global_sample_index

        current_entry = self._encoded[chunk_index - 1]  # type: ignore
        last_num_samples = current_entry[LAST_SEEN_INDEX_COLUMN] + 1

        return int(global_sample_index - last_num_samples)

    def _validate_incoming_item(self, _, num_samples: int):
        if num_samples < 0:
            raise ValueError(
                f"Cannot register negative num samples. Got: {num_samples}"
            )

        if self.num_samples == 0:
            raise ChunkIdEncoderError(
                "Cannot register samples because no chunk IDs exist."
            )

        if num_samples == 0 and self.num_chunks < 2:
            raise ChunkIdEncoderError(
                "Cannot register 0 num_samples (signifying a partial sample continuing the last chunk) when no last chunk exists."
            )

        # note: do not call super() method (num_samples can be 0)

    def _derive_next_last_index(self, last_index: ENCODING_DTYPE, num_samples: int):
        # this operation will trigger an overflow for the first addition, so supress the warning
        np.seterr(over="ignore")
        new_last_index = last_index + ENCODING_DTYPE(num_samples)
        np.seterr(over="warn")

        return new_last_index

    def _combine_condition(self, *args) -> bool:
        """Always returns True because sample registration can always be done. Used in base encoder `register_samples`."""

        return True

    def _derive_value(self, row: np.ndarray, *_) -> np.ndarray:
        return row[CHUNK_ID_COLUMN]

    def __setitem__(self, *args):
        raise NotImplementedError(
            "There is no reason for ChunkIdEncoder to be updated now."
        )



from typing import Optional
from hub.constants import AGREEMENT_FILENAME, HUB_CLOUD_DEV_USERNAME
from hub.core.dataset import Dataset
from hub.client.client import HubBackendClient
from hub.client.log import logger
from hub.util.agreement import handle_dataset_agreement
from hub.util.path import is_hub_cloud_path
from warnings import warn


class HubCloudDataset(Dataset):
    def __init__(self, path, *args, **kwargs):
        self._client = None
        self.path = path
        self.org_id, self.ds_name = None, None
        self._set_org_and_name()

        super().__init__(*args, **kwargs)

        if self.is_actually_cloud:
            handle_dataset_agreement(self.agreement, path, self.ds_name, self.org_id)
        else:
            # NOTE: this can happen if you override `hub.core.dataset.FORCE_CLASS`
            warn(
                f'Created a hub cloud dataset @ "{self.path}" which does not have the "hub://" prefix. Note: this dataset should only be used for testing!'
            )

    @property
    def client(self):
        if self._client is None:
            self._client = HubBackendClient(token=self._token)
        return self._client

    @property
    def is_actually_cloud(self) -> bool:
        """Datasets that are connected to hub cloud can still technically be stored anywhere.
        If a dataset is hub cloud but stored without `hub://` prefix, it should only be used for testing.
        """

        return is_hub_cloud_path(self.path)

    @property
    def token(self):
        """Get attached token of the dataset"""
        if self._token is None:
            self._token = self.client.get_token()
        return self._token

    def _set_org_and_name(self):
        if self.is_actually_cloud:
            split_path = self.path.split("/")
            self.org_id, self.ds_name = split_path[2], split_path[3]
        else:
            # if this dataset isn't actually pointing to a datset in the cloud
            # a.k.a this dataset is trying to simulate a hub cloud dataset
            # it's safe to assume they want to use the dev org
            self.org_id = HUB_CLOUD_DEV_USERNAME
            self.ds_name = self.path.replace("/", "_").replace(".", "")

    def _register_dataset(self):
        # called in super()._populate_meta

        self.client.create_dataset_entry(
            self.org_id,
            self.ds_name,
            self.version_state["meta"].__getstate__(),
            public=self.public,
        )

    def make_public(self):
        if not self.public:
            self.client.update_privacy(self.org_id, self.ds_name, public=True)
            self.public = True

    def make_private(self):
        if self.public:
            self.client.update_privacy(self.org_id, self.ds_name, public=False)
            self.public = False

    def delete(self, large_ok=False):
        super().delete(large_ok=large_ok)

        self.client.delete_dataset_entry(self.org_id, self.ds_name)
        logger.info(f"Hub Dataset {self.path} successfully deleted.")

    @property
    def agreement(self) -> Optional[str]:
        try:
            agreement_bytes = self.storage[AGREEMENT_FILENAME]
            return agreement_bytes.decode("utf-8")
        except KeyError:
            return None

    def add_agreeement(self, agreement: str):
        self.storage.check_readonly()
        self.storage[AGREEMENT_FILENAME] = agreement.encode("utf-8")



import pickle
from typing import Dict, Optional, Set
from hub.client.utils import get_user_name
from hub.client.config import ALL_AGREEMENTS_PATH
from hub.util.exceptions import AgreementNotAcceptedError, NotLoggedInError


def update_local_agreements(all_local_agreements: Dict[str, Set[str]]):
    with open(ALL_AGREEMENTS_PATH, "wb") as f:
        pickle.dump(all_local_agreements, f)


def get_all_local_agreements() -> Dict[str, Set[str]]:
    try:
        with open(ALL_AGREEMENTS_PATH, "rb") as f:
            return pickle.load(f)
    except FileNotFoundError:
        return {}


def agreement_prompt(org_id: str, ds_name: str, agreement: str):
    print()
    print()
    print(
        "The owner of the dataset you are trying to access requires that you agree to the following terms first:"
    )
    print()
    print("-" * 16)
    print(f"Dataset Owner: {org_id}")
    print(f"Dataset Name: {ds_name}")
    print("-" * 16)

    print("Terms:")
    print(agreement)
    print()
    print()

    print("-" * 16)
    user_input = input(
        f"In order to accept, please type the dataset's name ({ds_name}) and press enter: "
    )
    return user_input == ds_name


def handle_dataset_agreement(
    agreement: Optional[str], path: str, ds_name: str, org_id: str
):
    if agreement is None:
        return
    user_name = get_user_name()
    if user_name == "public":
        raise NotLoggedInError()
    if user_name == "org_id":
        return
    all_local_agreements = get_all_local_agreements()
    agreement_set = all_local_agreements.get(user_name) or set()
    if path not in agreement_set:
        accepted = agreement_prompt(org_id, ds_name, agreement)
        if accepted:
            print("Accepted agreement!")
            agreement_set.add(path)
            all_local_agreements[user_name] = agreement_set
            update_local_agreements(all_local_agreements)
        else:
            raise AgreementNotAcceptedError()



from hub.util.exceptions import ModuleNotInstalledException, UnsupportedSchedulerError
from hub.util.check_installation import ray_installed
from hub.core.compute.provider import ComputeProvider


def get_compute_provider(
    scheduler: str = "threaded", num_workers: int = 0
) -> ComputeProvider:
    num_workers = max(num_workers, 0)
    if scheduler == "serial" or num_workers == 0:
        from hub.core.compute.serial import SerialProvider

        compute: ComputeProvider = SerialProvider()
    elif scheduler == "threaded":
        from hub.core.compute.thread import ThreadProvider

        compute = ThreadProvider(num_workers)
    elif scheduler == "processed":
        from hub.core.compute.process import ProcessProvider

        compute = ProcessProvider(num_workers)
    elif scheduler == "ray":
        if not ray_installed():
            raise ModuleNotInstalledException(
                "'ray' should be installed to use ray scheduler."
            )
        from hub.core.compute.ray import RayProvider

        compute = RayProvider(num_workers)

    else:
        raise UnsupportedSchedulerError(scheduler)
    return compute



def test_dataset(local_ds_generator):
    ds = local_ds_generator()

    assert len(ds.info) == 0

    ds.info.update(my_key=0)
    ds.info.update(my_key=1)

    ds.info.update(another_key="hi")
    ds.info.update({"another_key": "hello"})

    ds.info.update({"something": "aaaaa"}, something="bbbb")

    ds.info.test = [1, 2, "5"]

    test_list = ds.info.test
    with ds:
        ds.info.update({"test2": (1, 5, (1, "2"), [5, 6, (7, 8)])})
        ds.info.update(xyz="abc")
        test_list.extend(["user made change without `update`"])

    ds.info.update({"1_-+": 5})

    ds = local_ds_generator()

    assert len(ds.info) == 7

    assert ds.info.another_key == "hello"
    assert ds.info.something == "bbbb"

    assert ds.info.test == [1, 2, "5", "user made change without `update`"]
    assert ds.info.test2 == [1, 5, [1, "2"], [5, 6, [7, 8]]]

    assert ds.info.xyz == "abc"
    assert ds.info["1_-+"] == 5  # key can't be accessed with `.` syntax

    ds.info.update(test=[99])

    ds = local_ds_generator()

    assert len(ds.info) == 7
    assert ds.info.test == [99]

    ds.info.delete("test")
    assert len(ds.info) == 6

    ds.info.delete(["1_-+", "xyz"])
    assert len(ds.info) == 4

    ds.info.delete()
    assert len(ds.info) == 0


def test_tensor(local_ds_generator):
    ds = local_ds_generator()

    t1 = ds.create_tensor("tensor1")
    t2 = ds.create_tensor("tensor2")

    t1.info.update(key=0)
    t2.info.update(key=1, key1=0)
    t2.info.key2 = 2
    t2.info["key3"] = 3

    ds = local_ds_generator()

    t1 = ds.tensor1
    t2 = ds.tensor2

    assert len(t1.info) == 1
    assert len(t2.info) == 4

    assert t1.info.key == t1.info["key"] == 0
    assert t2.info.key == t2.info["key"] == 1
    assert t2.info.key1 == t2.info["key1"] == 0
    assert t2.info.key2 == t2.info["key2"] == 2
    assert t2.info.key3 == t2.info["key3"] == 3

    with ds:
        t1.info.update(key=99)

    ds = local_ds_generator()

    t1 = ds.tensor1
    t2 = ds.tensor2

    assert len(t1.info) == 1
    assert len(t2.info) == 4

    assert t1.info.key == 99

    t2.info.delete("key")
    assert len(t2.info) == 3

    t2.info.delete(["key2", "key3"])
    assert len(t2.info) == 1

    t2.info.delete()
    assert len(t2.info) == 0


def test_update_reference_manually(local_ds_generator):
    """Right now synchronization can only happen when you call `info.update`."""

    ds = local_ds_generator()

    ds.info.update(key=[1, 2, 3])

    ds = local_ds_generator()

    l = ds.info.key
    assert l == [1, 2, 3]

    # un-registered update
    l.append(5)
    assert ds.info.key == [1, 2, 3, 5]

    ds = local_ds_generator()

    l = ds.info.key
    assert l == [1, 2, 3]

    # registered update
    l.append(99)
    ds.info.update()

    ds = local_ds_generator()

    assert l == [1, 2, 3, 99]


def test_class_label(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("labels", htype="class_label", class_names=["a", "b", "c"])
    ds.create_tensor("labels2", htype="class_label")
    assert len(ds.labels.info) == 1
    assert len(ds.labels2.info) == 1
    assert (
        ds.labels.info.class_names == ds.labels.info["class_names"] == ["a", "b", "c"]
    )
    assert ds.labels2.info.class_names == ds.labels2.info["class_names"] == []
    ds.labels.info.class_names = ["c", "b", "a"]
    ds = local_ds_generator()
    assert len(ds.labels.info) == 1
    assert len(ds.labels2.info) == 1
    assert (
        ds.labels.info.class_names == ds.labels.info["class_names"] == ["c", "b", "a"]
    )
    assert ds.labels2.info.class_names == ds.labels2.info["class_names"] == []



import hub
from hub.util.exceptions import (
    SampleCompressionError,
    SampleDecompressionError,
    UnsupportedCompressionError,
    CorruptedSampleError,
)
from hub.compression import (
    get_compression_type,
    BYTE_COMPRESSION,
    IMAGE_COMPRESSION,
    VIDEO_COMPRESSION,
    AUDIO_COMPRESSION,
)
from typing import Union, Tuple, Sequence, List, Optional, BinaryIO
import numpy as np

from PIL import Image, UnidentifiedImageError  # type: ignore
from io import BytesIO
import mmap
import struct
import sys
import re
import numcodecs.lz4  # type: ignore
import lz4.frame  # type: ignore
import os
import subprocess as sp
import tempfile
from miniaudio import (  # type: ignore
    mp3_read_file_f32,
    mp3_read_f32,
    mp3_get_file_info,
    mp3_get_info,
    flac_read_file_f32,
    flac_read_f32,
    flac_get_file_info,
    flac_get_info,
    wav_read_file_f32,
    wav_read_f32,
    wav_get_file_info,
    wav_get_info,
)
from numpy.core.fromnumeric import compress  # type: ignore
import math


if sys.byteorder == "little":
    _NATIVE_INT32 = "<i4"
    _NATIVE_FLOAT32 = "<f4"
else:
    _NATIVE_INT32 = ">i4"
    _NATIVE_FLOAT32 = ">f4"

if os.name == "nt":
    _FFMPEG_BINARY = "ffmpeg.exe"
    _FFPROBE_BINARY = "ffprobe.exe"
else:
    _FFMPEG_BINARY = "ffmpeg"
    _FFPROBE_BINARY = "ffprobe"

DIMS_RE = re.compile(rb" ([0-9]+)x([0-9]+)")
FPS_RE = re.compile(rb" ([0-9]+) fps,")
DURATION_RE = re.compile(rb"Duration: ([0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]{2}),")
INFO_RE = re.compile(rb"([a-z]+)=([0-9./]+)")

_JPEG_SOFS = [
    b"\xff\xc0",
    b"\xff\xc2",
    b"\xff\xc1",
    b"\xff\xc3",
    b"\xff\xc5",
    b"\xff\xc6",
    b"\xff\xc7",
    b"\xff\xc9",
    b"\xff\xca",
    b"\xff\xcb",
    b"\xff\xcd",
    b"\xff\xce",
    b"\xff\xcf",
    b"\xff\xde",
    # Skip:
    b"\xff\xcc",
    b"\xff\xdc",
    b"\xff\xdd",
    b"\xff\xdf",
    # App: (0xFFE0 - 0xFFEF):
    *map(lambda x: x.to_bytes(2, "big"), range(0xFFE0, 0xFFF0)),
    # DQT:
    b"\xff\xdb",
    # COM:
    b"\xff\xfe",
    # Start of scan
    b"\xff\xda",
]

_JPEG_SKIP_MARKERS = set(_JPEG_SOFS[14:])
_JPEG_SOFS_RE = re.compile(b"|".join(_JPEG_SOFS))
_STRUCT_HHB = struct.Struct(">HHB")
_STRUCT_II = struct.Struct(">ii")

_HUB_MKV_HEADER = b"HUB_MKV_META"

_FFMPEG_EXISTS = None


def ffmpeg_exists():
    global _FFMPEG_EXISTS
    if _FFMPEG_EXISTS is None:
        _FFMPEG_EXISTS = True
        try:
            retval = sp.run(
                [_FFMPEG_BINARY, "-h"], stdout=sp.PIPE, stderr=sp.PIPE
            ).returncode
        except FileNotFoundError as e:
            _FFMPEG_EXISTS = False
    return _FFMPEG_EXISTS


def ffmpeg_binary():
    if ffmpeg_exists():
        return _FFMPEG_BINARY
    raise FileNotFoundError(
        "FFMPEG not found. Install FFMPEG to use hub's video features"
    )


def ffprobe_binary():
    if ffmpeg_exists():
        return _FFPROBE_BINARY
    raise FileNotFoundError(
        "FFMPEG not found. Install FFMPEG to use hub's video features"
    )


def to_image(array: np.ndarray) -> Image:
    shape = array.shape
    if len(shape) == 3 and shape[0] != 1 and shape[2] == 1:
        # convert (X,Y,1) grayscale to (X,Y) for pillow compatibility
        return Image.fromarray(array.squeeze(axis=2))

    return Image.fromarray(array)


def _compress_apng(array: np.ndarray) -> bytes:
    if array.ndim == 3:
        # Binary APNG
        frames = list(
            map(Image.fromarray, (array[:, :, i] for i in range(array.shape[2])))
        )
    elif array.ndim == 4 and array.shape[3] <= 4:
        # RGB(A) APNG
        frames = list(map(Image.fromarray, array))
    else:
        raise SampleCompressionError(array.shape, "apng", "Unexpected shape.")
    out = BytesIO()
    frames[0].save(out, "png", save_all=True, append_images=frames[1:])
    out.seek(0)
    ret = out.read()
    out.close()
    return ret


def _decompress_apng(buffer: Union[bytes, memoryview]) -> np.ndarray:
    img = Image.open(BytesIO(buffer))
    frame0 = np.array(img)
    if frame0.ndim == 2:
        ret = np.zeros(frame0.shape + (img.n_frames,), dtype=frame0.dtype)
        ret[:, :, 0] = frame0
        for i in range(1, img.n_frames):
            img.seek(i)
            ret[:, :, i] = np.array(img)
    else:
        ret = np.zeros((img.n_frames,) + frame0.shape, dtype=frame0.dtype)
        ret[0] = frame0
        for i in range(1, img.n_frames):
            img.seek(i)
            ret[i] = np.array(img)
    return ret


def compress_bytes(buffer: Union[bytes, memoryview], compression: str) -> bytes:
    if compression == "lz4":
        return numcodecs.lz4.compress(buffer)
    else:
        raise SampleCompressionError(
            (len(buffer),), compression, f"Not a byte compression: {compression}"
        )


def decompress_bytes(buffer: Union[bytes, memoryview], compression: str) -> bytes:
    if not buffer:
        return b""
    if compression == "lz4":
        if (
            buffer[:4] == b'\x04"M\x18'
        ):  # python-lz4 magic number (backward compatiblity)
            return lz4.frame.decompress(buffer)
        return numcodecs.lz4.decompress(buffer)
    else:
        raise SampleDecompressionError()


def compress_array(array: np.ndarray, compression: str) -> bytes:
    """Compress some numpy array using `compression`. All meta information will be contained in the returned buffer.

    Note:
        `decompress_array` may be used to decompress from the returned bytes back into the `array`.

    Args:
        array (np.ndarray): Array to be compressed.
        compression (str): `array` will be compressed with this compression into bytes. Right now only arrays compatible with `PIL` will be compressed.

    Raises:
        UnsupportedCompressionError: If `compression` is unsupported. See `hub.compressions`.
        SampleCompressionError: If there was a problem compressing `array`.
        NotImplementedError: If compression is not supported.

    Returns:
        bytes: Compressed `array` represented as bytes.
    """

    # empty sample shouldn't be compressed

    if 0 in array.shape:
        return bytes()

    if compression not in hub.compressions:
        raise UnsupportedCompressionError(compression)

    if compression is None:
        return array.tobytes()

    compr_type = get_compression_type(compression)

    if compr_type == BYTE_COMPRESSION:
        return compress_bytes(array.tobytes(), compression)
    elif compr_type == AUDIO_COMPRESSION:
        raise NotImplementedError(
            "In order to store audio data, you should use `hub.read(path_to_file)`. Compressing raw data is not yet supported."
        )
    elif compr_type == VIDEO_COMPRESSION:
        raise NotImplementedError(
            "In order to store video data, you should use `hub.read(path_to_file)`. Compressing raw data is not yet supported."
        )

    if compression == "apng":
        return _compress_apng(array)
    try:
        img = to_image(array)
        out = BytesIO()
        out._close = out.close  # type: ignore
        out.close = (  # type: ignore
            lambda: None
        )  # sgi save handler will try to close the stream (see https://github.com/python-pillow/Pillow/pull/5645)
        kwargs = {"sizes": [img.size]} if compression == "ico" else {}
        img.save(out, compression, **kwargs)
        out.seek(0)
        compressed_bytes = out.read()
        out._close()  # type: ignore
        decompress_array(compressed_bytes, array.shape)
        return compressed_bytes
    except (TypeError, OSError) as e:
        raise SampleCompressionError(array.shape, compression, str(e))


def decompress_array(
    buffer: Union[bytes, memoryview, str],
    shape: Optional[Tuple[int, ...]] = None,
    dtype: Optional[str] = None,
    compression: Optional[str] = None,
) -> np.ndarray:
    """Decompress some buffer into a numpy array. It is expected that all meta information is
    stored inside `buffer`.

    Note:
        `compress_array` may be used to get the `buffer` input.

    Args:
        buffer (bytes, memoryview, str): Buffer or file to be decompressed. It is assumed all meta information required to
            decompress is contained within `buffer`, except for byte compressions
        shape (Tuple[int], Optional): Desired shape of decompressed object. Reshape will attempt to match this shape before returning.
        dtype (str, Optional): Applicable only for byte compressions. Expected dtype of decompressed array.
        compression (str, Optional): Applicable only for byte compressions. Compression used to compression the given buffer.

    Raises:
        SampleDecompressionError: If decompression fails.
        ValueError: If dtype and shape are not specified for byte compression.

    Returns:
        np.ndarray: Array from the decompressed buffer.
    """
    compr_type = get_compression_type(compression)
    if compr_type == BYTE_COMPRESSION:
        if dtype is None or shape is None:
            raise ValueError("dtype and shape must be specified for byte compressions.")
        try:
            decompressed_bytes = decompress_bytes(buffer, compression)  # type: ignore
            return np.frombuffer(decompressed_bytes, dtype=dtype).reshape(shape)
        except Exception:
            raise SampleDecompressionError()
    elif compr_type == AUDIO_COMPRESSION:
        return _decompress_audio(buffer, compression)
    elif compr_type == VIDEO_COMPRESSION:
        return _decompress_video(buffer)

    if compression == "apng":
        return _decompress_apng(buffer)  # type: ignore
    try:
        if not isinstance(buffer, str):
            buffer = BytesIO(buffer)  # type: ignore
        img = Image.open(buffer)  # type: ignore
        arr = np.array(img)
        if shape is not None:
            arr = arr.reshape(shape)
        return arr
    except Exception:
        raise SampleDecompressionError()


def _get_bounding_shape(shapes: Sequence[Tuple[int, ...]]) -> Tuple[int, int, int]:
    """Gets the shape of a bounding box that can contain the given the shapes tiled horizontally."""
    if len(shapes) == 0:
        return (0, 0, 0)
    channels_shape = shapes[0][2:]
    for shape in shapes:
        if shape[2:] != channels_shape:
            raise ValueError()
    return (max(s[0] for s in shapes), sum(s[1] for s in shapes)) + channels_shape  # type: ignore


def compress_multiple(arrays: Sequence[np.ndarray], compression: str) -> bytes:
    """Compress multiple arrays of different shapes into a single buffer. Used for chunk wise compression.
    The arrays are tiled horizontally and padded with zeros to fit in a bounding box, which is then compressed."""
    dtype = arrays[0].dtype
    for arr in arrays:
        if arr.dtype != dtype:
            raise SampleCompressionError(
                [arr.shape for shape in arr],  # type: ignore
                compression,
                message="All arrays expected to have same dtype.",
            )
    compr_type = get_compression_type(compression)
    if compr_type == BYTE_COMPRESSION:
        return compress_bytes(
            b"".join(arr.tobytes() for arr in arrays), compression
        )  # Note: shape and dtype info not included
    elif compr_type == AUDIO_COMPRESSION:
        raise NotImplementedError("compress_multiple does not support audio samples.")
    elif compr_type == VIDEO_COMPRESSION:
        raise NotImplementedError("compress_multiple does not support video samples.")
    elif compression == "apng":
        raise NotImplementedError("compress_multiple does not support apng samples.")
    canvas = np.zeros(_get_bounding_shape([arr.shape for arr in arrays]), dtype=dtype)
    next_x = 0
    for arr in arrays:
        canvas[: arr.shape[0], next_x : next_x + arr.shape[1]] = arr
        next_x += arr.shape[1]
    return compress_array(canvas, compression=compression)


def decompress_multiple(
    buffer: Union[bytes, memoryview],
    shapes: Sequence[Tuple[int, ...]],
    dtype: Optional[Union[np.dtype, str]] = None,
    compression: Optional[str] = None,
) -> List[np.ndarray]:
    """Unpack a compressed buffer into multiple arrays."""
    if compression and get_compression_type(compression) == "byte":
        decompressed_buffer = memoryview(decompress_bytes(buffer, compression))
        arrays = []
        itemsize = np.dtype(dtype).itemsize
        for shape in shapes:
            nbytes = int(np.prod(shape) * itemsize)
            arrays.append(
                np.frombuffer(decompressed_buffer[:nbytes], dtype=dtype).reshape(shape)
            )
            decompressed_buffer = decompressed_buffer[nbytes:]
        return arrays
    canvas = decompress_array(buffer)
    arrays = []
    next_x = 0
    for shape in shapes:
        arrays.append(canvas[: shape[0], next_x : next_x + shape[1]])
        next_x += shape[1]
    return arrays


def verify_compressed_file(
    file: Union[str, BinaryIO, bytes, memoryview], compression: str
) -> Union[Tuple[Tuple[int, ...], str], None]:
    """Verify the contents of an image file
    Args:
        file (Union[str, BinaryIO, bytes, memoryview]): Path to the file or file like object or contents of the file
        compression (str): Expected compression of the image file
    """
    if isinstance(file, str):
        file = open(file, "rb")
        close = True
    elif hasattr(file, "read"):
        close = False
        file.seek(0)  # type: ignore
    else:
        close = False
    try:
        if compression == "png":
            return _verify_png(file)
        elif compression == "jpeg":
            return _verify_jpeg(file), "|u1"
        elif get_compression_type(compression) == AUDIO_COMPRESSION:
            return _read_audio_shape(file, compression), "<f4"  # type: ignore
        elif compression in ("mp4", "mkv", "avi"):
            if isinstance(file, (bytes, memoryview, str)):
                return _read_video_shape(file), "|u1"
        else:
            return _fast_decompress(file)
    except Exception as e:
        raise CorruptedSampleError(compression)
    finally:
        if close:
            file.close()  # type: ignore

    return None


def get_compression(header=None, path=None):
    if path:
        # These formats are recognized by file extension for now
        file_formats = ["mp3", "flac", "wav", "mp4", "mkv", "avi"]
        for fmt in file_formats:
            if str(path).lower().endswith("." + fmt):
                return fmt
    if header:
        if not Image.OPEN:
            Image.init()
        for fmt in Image.OPEN:
            accept = Image.OPEN[fmt][1]
            if accept and accept(header):
                return fmt.lower()
        raise SampleDecompressionError()


def _verify_png(buf):
    if not hasattr(buf, "read"):
        buf = BytesIO(buf)
    img = Image.open(buf)
    img.verify()
    return Image._conv_type_shape(img)


def _verify_jpeg(f):
    if hasattr(f, "read"):
        return _verify_jpeg_file(f)
    return _verify_jpeg_buffer(f)


def _verify_jpeg_buffer(buf: bytes):
    # Start of Image
    mview = memoryview(buf)
    assert buf.startswith(b"\xff\xd8")
    # Look for Start of Frame
    sof_idx = -1
    offset = 0
    while True:
        match = _re_find_first(_JPEG_SOFS_RE, mview[offset:])
        if match is None:
            break
        idx = match.start(0) + offset
        marker = buf[idx : idx + 2]
        if marker == _JPEG_SOFS[-1]:
            break
        offset = idx + int.from_bytes(buf[idx + 2 : idx + 4], "big") + 2
        if marker not in _JPEG_SKIP_MARKERS:
            sof_idx = idx
    if sof_idx == -1:
        raise Exception()

    length = int.from_bytes(mview[sof_idx + 2 : sof_idx + 4], "big")
    assert mview[sof_idx + length + 2 : sof_idx + length + 4] in [
        b"\xff\xc4",
        b"\xff\xdb",
        b"\xff\xdd",
        b"\xff\xda",
    ]  # DHT, DQT, DRI, SOS
    shape = _STRUCT_HHB.unpack(mview[sof_idx + 5 : sof_idx + 10])
    assert buf.find(b"\xff\xd9") != -1
    if shape[-1] in (1, None):
        shape = shape[:-1]
    return shape


def _verify_jpeg_file(f):
    # See: https://dev.exiv2.org/projects/exiv2/wiki/The_Metadata_in_JPEG_files#2-The-metadata-structure-in-JPEG
    mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
    mv = memoryview(mm)
    try:
        soi = f.read(2)
        # Start of Image
        assert soi == b"\xff\xd8"

        # Look for Start of Frame
        sof_idx = -1
        offset = 0
        while True:
            view = mv[offset:]
            match = _re_find_first(_JPEG_SOFS_RE, view)
            view.release()
            if match is None:
                break
            idx = match.start(0) + offset
            marker = mm[idx : idx + 2]
            if marker == _JPEG_SOFS[-1]:
                break
            f.seek(idx + 2)
            offset = idx + int.from_bytes(f.read(2), "big") + 2
            if marker not in _JPEG_SKIP_MARKERS:
                sof_idx = idx
        if sof_idx == -1:
            raise Exception()  # Caught by verify_compressed_file()

        f.seek(sof_idx + 2)
        length = int.from_bytes(f.read(2), "big")
        f.seek(length - 2, 1)
        definition_start = f.read(2)
        assert definition_start in [
            b"\xff\xc4",
            b"\xff\xdb",
            b"\xff\xdd",
            b"\xff\xda",
        ]  # DHT, DQT, DRI, SOS
        f.seek(sof_idx + 5)
        shape = _STRUCT_HHB.unpack(f.read(5))
        # TODO this check is too slow
        assert mm.find(b"\xff\xd9") != -1  # End of Image
        if shape[-1] in (1, None):
            shape = shape[:-1]
        return shape
    finally:
        mv.release()
        mm.close()


def _fast_decompress(buf):
    if not hasattr(buf, "read"):
        buf = BytesIO(buf)
    img = Image.open(buf)
    img.load()
    if img.mode == 1:
        args = ("L",)
    else:
        args = (img.mode,)
    enc = Image._getencoder(img.mode, "raw", args)
    enc.setimage(img.im)
    bufsize = max(65536, img.size[0] * 4)
    while True:
        status, err_code, buf = enc.encode(
            bufsize
        )  # See https://github.com/python-pillow/Pillow/blob/master/src/encode.c#L144
        if err_code:
            break
    if err_code < 0:
        raise Exception()  # caught by verify_compressed_file()
    return Image._conv_type_shape(img)


def read_meta_from_compressed_file(
    file, compression: Optional[str] = None
) -> Tuple[str, Tuple[int], str]:
    """Reads shape, dtype and format without decompressing or verifying the sample."""
    if isinstance(file, str):
        f = open(file, "rb")
        isfile = True
        close = True
    elif hasattr(file, "read"):
        f = file
        close = False
        isfile = True
        f.seek(0)
    else:
        isfile = False
        f = file
        close = False
    try:
        if compression is None:
            path = file if isinstance(file, str) else None
            if hasattr(f, "read"):
                compression = get_compression(f.read(32), path)
                f.seek(0)
            else:
                compression = get_compression(f[:32], path)  # type: ignore
        if compression == "jpeg":
            try:
                shape, typestr = _read_jpeg_shape(f), "|u1"
            except Exception:
                raise CorruptedSampleError("jpeg")
        elif compression == "png":
            try:
                shape, typestr = _read_png_shape_and_dtype(f)
            except Exception:
                raise CorruptedSampleError("png")
        elif get_compression_type(compression) == AUDIO_COMPRESSION:
            try:
                shape, typestr = _read_audio_shape(file, compression), "<f4"
            except Exception as e:
                raise CorruptedSampleError(compression)
        elif compression in ("mp4", "mkv", "avi"):
            try:
                shape, typestr = _read_video_shape(file), "|u1"
            except Exception as e:
                raise CorruptedSampleError(compression)
        else:
            img = Image.open(f) if isfile else Image.open(BytesIO(f))  # type: ignore
            shape, typestr = Image._conv_type_shape(img)
            compression = img.format.lower()
        return compression, shape, typestr  # type: ignore
    finally:
        if close:
            f.close()


def _read_jpeg_shape(f: Union[bytes, BinaryIO]) -> Tuple[int, ...]:
    if hasattr(f, "read"):
        return _read_jpeg_shape_from_file(f)
    return _read_jpeg_shape_from_buffer(f)  # type: ignore


def _re_find_first(pattern, string):
    for match in re.finditer(pattern, string):
        return match


def _read_jpeg_shape_from_file(f) -> Tuple[int, ...]:
    """Reads shape of a jpeg image from file without loading the whole image in memory"""
    mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_COPY)
    mv = memoryview(mm)
    try:
        # Look for Start of Frame
        sof_idx = -1
        offset = 0
        while True:
            view = mv[offset:]
            match = _re_find_first(_JPEG_SOFS_RE, view)
            view.release()
            if match is None:
                break
            idx = match.start(0) + offset
            marker = mm[idx : idx + 2]
            if marker == _JPEG_SOFS[-1]:
                break
            f.seek(idx + 2)
            offset = idx + int.from_bytes(f.read(2), "big") + 2
            if marker not in _JPEG_SKIP_MARKERS:
                sof_idx = idx
        if sof_idx == -1:
            raise Exception()
        f.seek(sof_idx + 5)
        shape = _STRUCT_HHB.unpack(f.read(5))  # type: ignore
        if shape[-1] in (1, None):
            shape = shape[:-1]
        return shape
    finally:
        mv.release()
        mm.close()


def _read_jpeg_shape_from_buffer(buf: bytes) -> Tuple[int, ...]:
    """Gets shape of a jpeg file from its contents"""
    # Look for Start of Frame
    mv = memoryview(buf)
    sof_idx = -1
    offset = 0
    while True:
        match = _re_find_first(_JPEG_SOFS_RE, mv[offset:])
        if match is None:
            break
        idx = match.start(0) + offset
        marker = buf[idx : idx + 2]
        if marker == _JPEG_SOFS[-1]:
            break
        offset = idx + int.from_bytes(buf[idx + 2 : idx + 4], "big") + 2
        if marker not in _JPEG_SKIP_MARKERS:
            sof_idx = idx
    if sof_idx == -1:
        raise Exception()
    shape = _STRUCT_HHB.unpack(memoryview(buf)[sof_idx + 5 : sof_idx + 10])  # type: ignore
    if shape[-1] in (1, None):
        shape = shape[:-1]
    return shape


def _read_png_shape_and_dtype(f: Union[bytes, BinaryIO]) -> Tuple[Tuple[int, ...], str]:
    """Reads shape and dtype of a png file from a file like object or file contents.
    If a file like object is provided, all of its contents are NOT loaded into memory."""
    if not hasattr(f, "read"):
        f = BytesIO(f)  # type: ignore
    f.seek(16)  # type: ignore
    size = _STRUCT_II.unpack(f.read(8))[::-1]  # type: ignore
    bits, colors = f.read(2)  # type: ignore

    # Get the number of channels and dtype based on bits and colors:
    if colors == 0:
        if bits == 1:
            typstr = "|b1"
        elif bits == 16:
            typstr = _NATIVE_INT32
        else:
            typstr = "|u1"
        nlayers = None
    else:
        typstr = "|u1"
        if colors == 2:
            nlayers = 3
        elif colors == 3:
            nlayers = None
        elif colors == 4:
            if bits == 8:
                nlayers = 2
            else:
                nlayers = 4
        else:
            nlayers = 4
    shape = size if nlayers is None else size + (nlayers,)
    return shape, typstr  # type: ignore


def _decompress_audio(
    file: Union[bytes, memoryview, str], compression: Optional[str]
) -> np.ndarray:
    decompressor = globals()[
        f"{compression}_read{'_file' if isinstance(file, str) else ''}_f32"
    ]
    if isinstance(file, memoryview):
        if (
            isinstance(file.obj, bytes)
            and file.strides == (1,)
            and file.shape == (len(file.obj),)
        ):
            file = file.obj
        else:
            file = bytes(file)
    raw_audio = decompressor(file)
    return np.frombuffer(raw_audio.samples, dtype="<f4").reshape(
        raw_audio.num_frames, raw_audio.nchannels
    )


def _read_audio_shape(
    file: Union[bytes, memoryview, str], compression: str
) -> Tuple[int, ...]:
    f_info = globals()[
        f"{compression}_get{'_file' if isinstance(file, str) else ''}_info"
    ]
    info = f_info(file)
    return (info.num_frames, info.nchannels)


def _strip_hub_mp4_header(buffer: bytes):
    if buffer[: len(_HUB_MKV_HEADER)] == _HUB_MKV_HEADER:
        return memoryview(buffer)[len(_HUB_MKV_HEADER) + 6 :]
    return buffer


def _decompress_video(
    file: Union[bytes, memoryview, str],
) -> np.ndarray:

    shape = _read_video_shape(file)

    command = [
        ffmpeg_binary(),
        "-i",
        "pipe:",
        "-f",
        "image2pipe",
        "-pix_fmt",
        "rgb24",
        "-vcodec",
        "rawvideo",
        "-",
    ]
    if isinstance(file, str):
        command[2] = file
        pipe = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE, bufsize=10 ** 8)
        raw_video = pipe.communicate()[0]
    else:
        file = _strip_hub_mp4_header(file)
        pipe = sp.Popen(
            command, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, bufsize=10 ** 8
        )
        raw_video = pipe.communicate(input=file)[0]  # type: ignore
    return np.frombuffer(raw_video[: int(np.prod(shape))], dtype=np.uint8).reshape(
        shape
    )


def _read_video_shape(file: Union[bytes, memoryview, str]) -> Tuple[int, ...]:
    info = _get_video_info(file)
    if info["duration"] is None:
        nframes = -1
    else:
        nframes = math.floor(info["duration"] * info["rate"])
    return (nframes, info["height"], info["width"], 3)


def _get_video_info(file: Union[bytes, memoryview, str]) -> dict:
    duration = None
    command = [
        ffprobe_binary(),
        "-select_streams",
        "v:0",
        "-show_entries",
        "stream=width,height,duration,r_frame_rate",
        "-of",
        "default=noprint_wrappers=1",
        "pipe:",
    ]

    if isinstance(file, str):
        command[-1] = file
        pipe = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE, bufsize=10 ** 5)
        raw_info = pipe.stdout.read()  # type: ignore
        raw_err = pipe.stderr.read()  # type: ignore
        pipe.communicate()
        duration = bytes.decode(re.search(DURATION_RE, raw_err).groups()[0])  # type: ignore
        duration = to_seconds(duration)
    else:
        if file[: len(_HUB_MKV_HEADER)] == _HUB_MKV_HEADER:
            mv = memoryview(file)
            n = len(_HUB_MKV_HEADER) + 2
            duration = struct.unpack("f", mv[n : n + 4])[0]
            file = mv[n + 4 :]
        pipe = sp.Popen(
            command, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, bufsize=10 ** 5
        )
        raw_info = pipe.communicate(input=file)[0]
    ret = dict(
        map(lambda kv: (bytes.decode(kv[0]), kv[1]), re.findall(INFO_RE, raw_info))
    )
    ret["width"] = int(ret["width"])
    ret["height"] = int(ret["height"])
    if "duration" in ret:
        ret["duration"] = float(ret["duration"])
    else:
        ret["duration"] = duration
    ret["rate"] = float(eval(ret["rate"]))
    return ret


DURATION_RE = re.compile(rb"Duration: ([0-9:.]+),")


def to_seconds(time):
    return sum([60 ** i * float(j) for (i, j) in enumerate(time.split(":")[::-1])])


def _to_hub_mkv(file: str):
    command = [
        ffmpeg_binary(),
        "-i",
        file,
        "-codec",
        "copy",
        "-f",
        "matroska",
        "pipe:",
    ]
    pipe = sp.Popen(
        command, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, bufsize=10 ** 5
    )
    mkv, raw_info = pipe.communicate()
    duration = bytes.decode(re.search(DURATION_RE, raw_info).groups()[0])  # type: ignore
    duration = to_seconds(duration)
    mkv = _HUB_MKV_HEADER + struct.pack("<Hf", 4, duration) + mkv
    return mkv



from hub.util.version_control import auto_checkout
from hub.core.storage.lru_cache import LRUCache
from typing import Any, Dict, Optional, Union, Sequence
from hub.core.storage.cachable import CachableCallback, use_callback


class Info(CachableCallback):
    def __init__(self):
        """Contains **optional** key/values that datasets/tensors use for human-readability.
        See the `Meta` class for required key/values for datasets/tensors.

        Note:
            Since `Info` is rarely written to and mostly by the user, every modifier will call `cache[key] = self`.
            Must call `initialize_callback_location` before using any methods.
        """
        self._info = {}
        super().__init__()

    @property
    def nbytes(self):
        # TODO: optimize this
        return len(self.tobytes())

    @use_callback(check_only=True)
    def __len__(self):
        return len(self._info)

    @use_callback(check_only=True)
    def __getstate__(self) -> Dict[str, Any]:
        return self._info

    def __setstate__(self, state: Dict[str, Any]):
        self._info = state

    @use_callback()
    def update(self, *args, **kwargs):
        """Store optional dataset/tensor information. Will be accessible after loading your data from a new script!
        Inputs must be supported by JSON.


        Note:
            This method has the same functionality as `dict().update(...)` Reference: https://www.geeksforgeeks.org/python-dictionary-update-method/.
            A full list of supported value types can be found here: https://docs.python.org/3/library/json.html#json.JSONEncoder.

        Examples:
            Normal update usage:
                >>> ds.info
                {}
                >>> ds.info.update(key=0)
                >>> ds.info
                {"key": 0}
                >>> ds.info.update({"key1": 5, "key2": [1, 2, "test"]})
                >>> ds.info
                {"key": 0, "key1": 5, "key2": [1, 2, "test"]}

            Alternate update usage:
                >>> ds.info
                {}
                >>> ds.info.update(list=[1, 2, "apple"])
                >>> ds.info
                {"list": [1, 2, "apple"]}
                >>> l = ds.info.list
                >>> l
                [1, 2, "apple"]
                >>> l.append(5)
                >>> l
                [1, 2, "apple", 5]
                >>> ds.info.update()  # required to be persistent!

        """
        self._cache.check_readonly()
        if self._version_state is not None:
            auto_checkout(self._version_state, self._cache)
        self._info.update(*args, **kwargs)

    def __getattribute__(self, name: str) -> Any:
        """Allows access to info values using the `.` syntax. Example: `info.description`."""

        if name == "_info":
            return super().__getattribute__(name)
        if name in self._info:
            return self.__getitem__(name)
        return super().__getattribute__(name)

    def __getitem__(self, key: str):
        return self._info[key]

    def __str__(self):
        return self._info.__str__()

    def __repr__(self):
        return self._info.__repr__()

    @use_callback()
    def delete(self, key: Optional[Union[Sequence[str], str]] = None):
        """Deletes a key or list of keys. If no key(s) is passed, all keys are deleted."""
        self._cache.check_readonly()
        if self._version_state is not None:
            auto_checkout(self._version_state, self._cache)
        if key is None:
            self._info.clear()
        elif isinstance(key, str):
            del self._info[key]
        elif isinstance(key, Sequence):
            for k in key:
                del self._info[k]
        else:
            raise KeyError(key)

    @use_callback()
    def __setitem__(self, key: str, value):
        self._cache.check_readonly()
        if self._version_state is not None:
            auto_checkout(self._version_state, self._cache)
        self._info[key] = value

    def __setattr__(self, key: str, value):
        if key in {"_key", "_cache", "_info", "_version_state"}:
            object.__setattr__(self, key, value)
        else:
            self[key] = value

    def __getattr__(self, key: str):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            if key == "_info":
                self._info = {}
                return self._info
            return self[key]

    # the below methods are used by cloudpickle dumps
    def __origin__(self):
        return None

    def __values__(self):
        return None

    def __type__(self):
        return None

    def __union_params__(self):
        return None

    def __tuple_params__(self):
        return None

    def __result__(self):
        return None

    def __args__(self):
        return None


def load_info(info_key: str, cache: LRUCache, version_state: Dict[str, Any]):
    if info_key in cache:
        info = cache.get_cachable(info_key, Info)
    else:
        info = Info()
        info.initialize_callback_location(info_key, cache, version_state)

    return info



from typing import List, Any
from random import randrange
from functools import reduce
from operator import mul
import warnings


class ShuffleBuffer:
    """Shuffling buffer used to shuffle samples by the rule:

    Given new sample if buffer is not full, add sample to the buffer else pick other sample
    randomly and swap it with given sample.

    Args:
        size(int):  size of the buffer in bytes
    Raises:
        ValueError if buffer size is not set
    """

    def __init__(self, size: int) -> None:
        if size <= 0:
            raise ValueError("Buffer size should be positive value more than zero")

        self.size = size
        self.buffer: List[Any] = list()
        self.buffer_used = 0

    def exchange(self, sample):
        """Shuffle with existing elements in a buffer and return value if buffer is full or if `None` is provided as argument.

        Args:
            sample: new sample to add or None

        Returns:
            random sample or None,
            same sample if buffer is empty and sample doesn't fit
        """
        buffer_len = len(self.buffer)

        if not sample is None:
            sample_size = self._sample_size(sample)

            # fill buffer of not reach limit
            if self.buffer_used + sample_size <= self.size:
                self.buffer_used += sample_size
                self.buffer.append(sample)
                return None

            if buffer_len == 0:
                warnings.warn(
                    f"Buffer size is too small. Sample with size {sample_size} does not fit in buffer of size {self.size}"
                )
                return sample

            # exchange samples with shuffle buffer
            selected = randrange(buffer_len)
            val = self.buffer[selected]
            self.buffer[selected] = sample

            self.buffer_used += sample_size
            self.buffer_used -= self._sample_size(val)
            return val
        else:
            if buffer_len > 0:

                # return random selection
                selected = randrange(buffer_len)
                val = self.buffer.pop(selected)
                self.buffer_used -= self._sample_size(val)

                return val
            else:
                return None

    def emtpy(self) -> bool:
        return len(self.buffer) == 0

    def _sample_size(self, sample):
        return sum(
            [
                tensor.storage().element_size() * reduce(mul, tensor.shape, 1)
                for _, tensor in sample.items()
            ]
        )

    def __len__(self):
        return len(self.buffer)

    def __str__(self) -> str:
        return f"ShuffleBuffer(size = {self.size}, buffer_used = {self.buffer_used}, samples = {len(self.buffer)})"



import pytest
import hub
from hub.util.exceptions import TensorInvalidSampleShapeError

WARNING_STR = "Grayscale images will be reshaped"


@pytest.fixture(params=["jpeg"])
def hub_read_images(request, grayscale_image_paths, color_image_paths):
    gray_path = grayscale_image_paths[request.param]
    color_path = color_image_paths[request.param]
    yield request.param, hub.read(gray_path), hub.read(color_path)


def make_tensor_and_append(ds, htype, sample_compression, images):
    ds.create_tensor("images", htype=htype, sample_compression=sample_compression)
    for img in images:
        ds.images.append(img)


def make_tensor_and_extend(ds, htype, sample_compression, images):
    ds.create_tensor("images", htype=htype, sample_compression=sample_compression)
    ds.images.extend(images)


def test_append_grayscale_second(local_ds_generator, hub_read_images):
    "Append a hub.read color image first, then a grayscale"
    ds = local_ds_generator()
    imgtype, gray, color = hub_read_images
    with pytest.warns(UserWarning, match=WARNING_STR):
        make_tensor_and_append(ds, "image", imgtype, [color, gray])
    assert len(ds.images) == 2
    assert ds.images.meta.min_shape[-1] == 1
    assert ds.images.meta.max_shape[-1] == 3


def test_append_grayscale_second_many(local_ds_generator, hub_read_images):
    "Append a hub.read color image first, then a mix of color and gray."
    ds = local_ds_generator()
    imgtype, gray, color = hub_read_images
    with pytest.warns(UserWarning, match=WARNING_STR):
        make_tensor_and_append(
            ds, "image", imgtype, [color, color, gray, color, gray, color]
        )
    assert len(ds.images) == 6
    assert ds.images.meta.min_shape[-1] == 1
    assert ds.images.meta.max_shape[-1] == 3


def test_extend_grayscale_second(local_ds_generator, hub_read_images):
    "Extend a dataset with a list of color first, gray second."
    ds = local_ds_generator()
    imgtype, gray, color = hub_read_images
    with pytest.warns(UserWarning, match=WARNING_STR):
        make_tensor_and_extend(ds, "image", imgtype, [color, gray])
    assert len(ds.images) == 2
    assert ds.images.meta.min_shape[-1] == 1
    assert ds.images.meta.max_shape[-1] == 3


@pytest.mark.xfail(raises=TensorInvalidSampleShapeError, strict=True)
def test_append_grayscale_first(local_ds_generator, hub_read_images):
    "Append a gray first, color second."
    ds = local_ds_generator()
    imgtype, gray, color = hub_read_images
    make_tensor_and_append(ds, "image", imgtype, [gray, color])


def test_append_grayscale_second_generic_ds(local_ds_generator, hub_read_images):
    "Append with htype=generic, sample_compression=<valid image compression>."
    ds = local_ds_generator()
    imgtype, gray, color = hub_read_images
    with pytest.warns(UserWarning, match=WARNING_STR):
        make_tensor_and_append(ds, "generic", imgtype, [color, gray])


@pytest.mark.xfail(raises=TensorInvalidSampleShapeError, strict=True)
def test_append_grayscale_second_generic_ds_unspecified_comp(
    local_ds_generator, hub_read_images
):
    "Append with htype=generic and sample_compression=unspecified."
    ds = local_ds_generator()
    _, gray, color = hub_read_images
    make_tensor_and_append(ds, "generic", "unspecified", [color, gray])


def test_append_two_grayscale(local_ds_generator, hub_read_images):
    "Append two hub.read grayscale images.  There should be no warning."
    ds = local_ds_generator()
    imgtype, gray, _ = hub_read_images
    make_tensor_and_append(ds, "image", imgtype, [gray, gray])
    assert len(ds.images) == 2
    assert len(ds.images.meta.min_shape) == 2
    assert list(ds.images.meta.min_shape) == list(gray.shape)
    assert list(ds.images.meta.max_shape) == list(gray.shape)


def test_append_many_grayscale(local_ds_generator, hub_read_images):
    "Append two hub.read grayscale images."
    ds = local_ds_generator()
    imgtype, gray, _ = hub_read_images
    make_tensor_and_append(ds, "image", imgtype, [gray, gray, gray, gray])
    assert len(ds.images) == 4
    assert len(ds.images.meta.min_shape) == 2
    assert list(ds.images.meta.min_shape) == list(gray.shape)
    assert list(ds.images.meta.max_shape) == list(gray.shape)



from hub.tests.common import get_actual_compression_from_buffer, assert_images_close
import numpy as np
import pytest
import hub
import lz4.frame  # type: ignore
from hub.core.compression import (
    compress_array,
    decompress_array,
    compress_multiple,
    decompress_multiple,
    verify_compressed_file,
    decompress_bytes,
)
from hub.compression import (
    get_compression_type,
    BYTE_COMPRESSION,
    IMAGE_COMPRESSION,
    IMAGE_COMPRESSIONS,
    BYTE_COMPRESSIONS,
    AUDIO_COMPRESSIONS,
    VIDEO_COMPRESSIONS,
    SUPPORTED_COMPRESSIONS,
)
from hub.util.exceptions import CorruptedSampleError
from PIL import Image  # type: ignore


compressions = SUPPORTED_COMPRESSIONS[:]
compressions.remove(None)  # type: ignore
compressions.remove("wmf")  # driver has to be provided by user for wmf write support

image_compressions = IMAGE_COMPRESSIONS[:]
image_compressions.remove("wmf")
image_compressions.remove("apng")


@pytest.mark.parametrize("compression", image_compressions + BYTE_COMPRESSIONS)
def test_array(compression, compressed_image_paths):
    # TODO: check dtypes and no information loss
    compression_type = get_compression_type(compression)
    if compression_type == BYTE_COMPRESSION:
        array = np.random.randint(0, 10, (32, 32))
    elif compression_type == IMAGE_COMPRESSION:
        array = np.array(hub.read(compressed_image_paths[compression][0]))
    shape = array.shape
    compressed_buffer = compress_array(array, compression)
    if compression_type == BYTE_COMPRESSION:
        decompressed_array = decompress_array(
            compressed_buffer, shape=shape, dtype=array.dtype, compression=compression
        )
    else:
        assert get_actual_compression_from_buffer(compressed_buffer) == compression
        decompressed_array = decompress_array(compressed_buffer, shape=shape)
    if compression == "png" or compression_type == BYTE_COMPRESSION:
        np.testing.assert_array_equal(array, decompressed_array)
    else:
        assert_images_close(array, decompressed_array)


@pytest.mark.parametrize("compression", image_compressions + BYTE_COMPRESSIONS)
def test_multi_array(compression, compressed_image_paths):
    compression_type = get_compression_type(compression)
    if compression_type == IMAGE_COMPRESSION:
        img = Image.open(compressed_image_paths[compression][0])
        img2 = img.resize((img.size[0] // 2, img.size[1] // 2))
        img3 = img.resize((img.size[0] // 3, img.size[1] // 3))
        arrays = list(map(np.array, [img, img2, img3]))
        compressed_buffer = compress_multiple(arrays, compression)
        decompressed_arrays = decompress_multiple(
            compressed_buffer, [arr.shape for arr in arrays]
        )
    elif compression_type == BYTE_COMPRESSION:
        arrays = [np.random.randint(0, 10, (32, 32)) for _ in range(3)]
        compressed_buffer = compress_multiple(arrays, compression)
        decompressed_arrays = decompress_multiple(
            compressed_buffer, [(32, 32)] * 3, arrays[0].dtype, compression
        )

    for arr1, arr2 in zip(arrays, decompressed_arrays):
        if compression == "png" or compression_type == BYTE_COMPRESSION:
            np.testing.assert_array_equal(arr1, arr2)
        else:
            assert_images_close(arr1, arr2)


@pytest.mark.parametrize("compression", image_compressions)
def test_verify(compression, compressed_image_paths, corrupt_image_paths):
    for path in compressed_image_paths[compression]:
        sample = hub.read(path)
        sample_loaded = hub.read(path)
        sample_loaded.compressed_bytes(compression)
        sample_verified_and_loaded = hub.read(path, verify=True)
        sample_verified_and_loaded.compressed_bytes(compression)
        pil_image_shape = np.array(Image.open(path)).shape
        assert (
            sample.shape
            == sample_loaded.shape
            == sample_verified_and_loaded.shape
            == pil_image_shape
        ), (
            sample.shape,
            sample_loaded.shape,
            sample_verified_and_loaded.shape,
            pil_image_shape,
        )
        verify_compressed_file(path, compression)
        with open(path, "rb") as f:
            verify_compressed_file(f, compression)
    if compression in corrupt_image_paths:
        path = corrupt_image_paths[compression]
        sample = hub.read(path)
        sample.compressed_bytes(compression)
        Image.open(path)
        with pytest.raises(CorruptedSampleError):
            sample = hub.read(path, verify=True)
            sample.compressed_bytes(compression)
        with pytest.raises(CorruptedSampleError):
            verify_compressed_file(path, compression)
        with pytest.raises(CorruptedSampleError):
            with open(path, "rb") as f:
                verify_compressed_file(f, compression)
        with pytest.raises(CorruptedSampleError):
            with open(path, "rb") as f:
                verify_compressed_file(f.read(), compression)


def test_lz4_bc():
    inp = np.random.random((100, 100)).tobytes()
    compressed = lz4.frame.compress(inp)
    decompressed = decompress_bytes(compressed, "lz4")
    assert decompressed == inp


def test_lz4_empty():
    assert decompress_bytes(b"", "lz4") == b""


@pytest.mark.parametrize("compression", AUDIO_COMPRESSIONS)
def test_audio(compression, audio_paths):
    path = audio_paths[compression]
    sample = hub.read(path)
    arr = np.array(sample)
    assert arr.dtype == "float32"
    with open(path, "rb") as f:
        assert sample.compressed_bytes(compression) == f.read()


@pytest.mark.parametrize("compression", VIDEO_COMPRESSIONS)
def test_video(compression, video_paths):
    path = video_paths[compression]
    sample = hub.read(path)
    arr = np.array(sample)
    assert arr.shape[-1] == 3
    assert arr.dtype == "uint8"
    if compression not in ("mp4", "mkv"):
        with open(path, "rb") as f:
            assert sample.compressed_bytes(compression) == f.read()


def test_apng(memory_ds):
    ds = memory_ds

    arrays = {
        "binary": [
            np.random.randint(
                0, 256, (25, 50, np.random.randint(100, 200)), dtype=np.uint8
            )
            for _ in range(10)
        ],
        "rgb": [
            np.random.randint(
                0, 256, (np.random.randint(100, 200), 32, 64, 3), dtype=np.uint8
            )
            for _ in range(10)
        ],
        "rgba": [
            np.random.randint(
                0, 256, (np.random.randint(100, 200), 16, 32, 4), dtype=np.uint8
            )
            for _ in range(10)
        ],
    }
    for k, v in arrays.items():
        with ds:
            ds.create_tensor(k, htype="image", sample_compression="apng")
            ds[k].extend(v)
        for arr1, arr2 in zip(ds[k].numpy(aslist=True), v):
            np.testing.assert_array_equal(arr1, arr2)



from hub.api.dataset import Dataset
from hub.tests.common import get_dummy_data_path
from hub.util.auto import get_most_common_extension, ingestion_summary
from io import StringIO
import sys
import pytest


def test_most_common_extension():
    path = get_dummy_data_path("tests_auto/auto_compression")
    path_file = get_dummy_data_path("test_auto/auto_comrpession/jpeg/bird.jpeg")

    compression = get_most_common_extension(path)
    file_compression = get_most_common_extension(path_file)

    assert compression == "png"
    assert file_compression == "jpeg"


def test_ingestion_summary_clean():
    clean_path = get_dummy_data_path("tests_auto/ingestion_summary/class1")

    ingest_summary_clean = StringIO()
    sys.stdout = ingest_summary_clean
    ingestion_summary(clean_path, [])
    sys.stdout = sys.__stdout__
    output = ingest_summary_clean.getvalue()

    assert output == "\n\nIngestion Complete. No files were skipped.\n\n\n"


def test_ingestion_summary_skipped():
    skipped_path = get_dummy_data_path("tests_auto/ingestion_summary")

    ingest_summary_skipped = StringIO()
    sys.stdout = ingest_summary_skipped
    ingestion_summary(skipped_path, ["test.json", "test.json"])
    sys.stdout = sys.__stdout__
    output = ingest_summary_skipped.getvalue()

    assert (
        output
        == "\n\ningestion_summary/    \n      [Skipped]  test.json\n      class0/    \n            [Skipped]  test.json\n      class1/    \n"
    )



import hub
from typing import Any, Dict, List, Tuple, Optional
from json.decoder import JSONDecodeError
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.storage import StorageProvider, MemoryProvider, LRUCache
from hub.core.chunk_engine import ChunkEngine
from hub.core.meta.encode.chunk_id import ChunkIdEncoder
from hub.core.transform.transform_dataset import TransformDataset
from hub.core.ipc import Client


from hub.constants import MB, TRANSFORM_PROGRESSBAR_UPDATE_INTERVAL
from hub.util.remove_cache import get_base_storage
from hub.util.keys import get_tensor_meta_key
from hub.util.exceptions import (
    InvalidInputDataError,
    InvalidOutputDatasetError,
    InvalidTransformDataset,
    TensorMismatchError,
)

import posixpath
import time


def transform_sample(
    sample: Any,
    pipeline,
) -> TransformDataset:
    """Calls all the functions one after the other on a single sample.
    Can return 0 or more samples.

    Args:
        sample: The sample on which the pipeline of functions is to be applied.
        pipeline (Pipeline): The Sequence of functions to apply on the sample.

    Raises:
        InvalidTransformDataset: If number of tensors were inconsistent between all transform datasets.

    Returns:
        TransformDataset: A transform dataset containing all the samples that were generated.
    """

    result = sample
    for index in range(len(pipeline)):
        transform_fn = pipeline.functions[index]
        fn, args, kwargs = transform_fn.func, transform_fn.args, transform_fn.kwargs

        if isinstance(result, TransformDataset):
            all_samples_out = []
            for item in result:
                samples_out = TransformDataset()
                fn(item, samples_out, *args, **kwargs)
                validate_transform_dataset(samples_out)
                all_samples_out.append(samples_out)
            result = combine_transform_datasets(all_samples_out)
            try:
                validate_transform_dataset(result)
            except InvalidTransformDataset:
                raise InvalidTransformDataset(
                    "One or more of the TransformDatasets returned had different number of tensors. Always ensure that all outputs have exactly the same tensors and equal number of samples in each tensor."
                )
        else:
            samples_out = TransformDataset()
            fn(result, samples_out, *args, **kwargs)
            validate_transform_dataset(samples_out)
            result = samples_out
    return result


def combine_transform_datasets(datasets: List[TransformDataset]):
    """Combines multiple TransformDataset into a single transform dataset."""
    final_ds = TransformDataset()
    for ds in datasets:
        for tensor in ds.tensors:
            final_ds[tensor].extend(ds[tensor].numpy())
    return final_ds


def validate_transform_dataset(dataset: TransformDataset):
    """Checks if the length of all the tensors is equal. Raises exception if not equal."""
    lengths = [len(dataset[tensor]) for tensor in dataset.tensors]
    if any(length != lengths[0] for length in lengths):
        raise InvalidTransformDataset


def is_empty_transform_dataset(dataset: TransformDataset):
    """Checks if there is any data in the TransformDataset. Returns True if empty, False otherwise."""
    return all(len(dataset[tensor]) == 0 for tensor in dataset.tensors)


def store_data_slice(
    transform_input: Tuple,
) -> Tuple[Dict[str, TensorMeta], Dict[str, ChunkIdEncoder]]:
    """Takes a slice of the original data and iterates through it and stores it in the actual storage.
    The tensor_meta and chunk_id_encoder are not stored to the storage to prevent overwrites/race conditions b/w workers.
    They are instead stored in memory and returned."""
    (
        data_slice,
        (output_storage, group_index),
        tensors,
        pipeline,
        version_state,
        progress_port,
    ) = transform_input
    all_chunk_engines = create_worker_chunk_engines(
        tensors, output_storage, version_state
    )

    if isinstance(data_slice, hub.Dataset):
        data_slice = add_cache_to_dataset_slice(data_slice, tensors)

    transform_data_slice_and_append(
        data_slice, pipeline, tensors, all_chunk_engines, group_index, progress_port
    )

    # retrieve the tensor metas and chunk_id_encoder from the memory
    all_tensor_metas = {}
    all_chunk_id_encoders = {}
    for tensor, chunk_engine in all_chunk_engines.items():
        chunk_engine.cache.flush()
        chunk_engine.meta_cache.flush()
        all_tensor_metas[tensor] = chunk_engine.tensor_meta
        all_chunk_id_encoders[tensor] = chunk_engine.chunk_id_encoder
    return all_tensor_metas, all_chunk_id_encoders


def _transform_sample_and_update_chunk_engines(
    sample,
    pipeline,
    tensors: List[str],
    all_chunk_engines: Dict[str, ChunkEngine],
    group_index: str,
):
    result = transform_sample(sample, pipeline)
    if is_empty_transform_dataset(result):
        return
    result_resolved = {
        posixpath.join(group_index, k): result[k] for k in result.tensors
    }
    result = result_resolved  # type: ignore
    if set(result.keys()) != set(tensors):
        raise TensorMismatchError(list(tensors), list(result.keys()))
    for tensor, value in result.items():
        all_chunk_engines[tensor].extend(value.numpy_compressed())


def transform_data_slice_and_append(
    data_slice,
    pipeline,
    tensors: List[str],
    all_chunk_engines: Dict[str, ChunkEngine],
    group_index: str,
    progress_port: Optional[int] = None,
) -> None:
    """Transforms the data_slice with the pipeline and adds the resultant samples to chunk_engines."""

    if progress_port is not None:
        last_reported_time = time.time()
        last_reported_num_samples = 0
        report_interval = TRANSFORM_PROGRESSBAR_UPDATE_INTERVAL
        client = Client(progress_port)
    try:
        n = len(data_slice)
        for i, sample in enumerate(data_slice):
            _transform_sample_and_update_chunk_engines(
                sample, pipeline, tensors, all_chunk_engines, group_index
            )
            if progress_port is not None:
                curr_time = time.time()
                if curr_time - last_reported_time > report_interval or i == n - 1:
                    num_samples = i + 1
                    client.send(num_samples - last_reported_num_samples)
                    last_reported_num_samples = num_samples
                    last_reported_time = curr_time
    except Exception as e:
        if progress_port is not None:
            client.send(str(e))
        else:
            raise e
    finally:
        if progress_port is not None:
            client.close()


def create_worker_chunk_engines(
    tensors: List[str], output_storage: StorageProvider, version_state
) -> Dict[str, ChunkEngine]:
    """Creates chunk engines corresponding to each storage for all tensors.
    These are created separately for each worker for parallel uploads.
    """
    all_chunk_engines = {}
    num_tries = 1000
    for tensor in tensors:
        for i in range(num_tries):
            try:
                # TODO: replace this with simply a MemoryProvider once we get rid of cachable
                memory_cache = LRUCache(MemoryProvider(), MemoryProvider(), 32 * MB)
                memory_cache.autoflush = False
                storage_cache = LRUCache(MemoryProvider(), output_storage, 32 * MB)
                storage_cache.autoflush = False

                # this chunk engine is used to retrieve actual tensor meta and chunk_size

                storage_chunk_engine = ChunkEngine(tensor, storage_cache, version_state)
                existing_meta = storage_chunk_engine.tensor_meta
                chunk_size = storage_chunk_engine.max_chunk_size
                new_tensor_meta = TensorMeta(
                    htype=existing_meta.htype,
                    dtype=existing_meta.dtype,
                    sample_compression=existing_meta.sample_compression,
                    chunk_compression=existing_meta.chunk_compression,
                    max_chunk_size=chunk_size,
                )
                meta_key = get_tensor_meta_key(tensor, version_state["commit_id"])
                memory_cache[meta_key] = new_tensor_meta  # type: ignore
                storage_cache.clear_cache()
                storage_chunk_engine = ChunkEngine(
                    tensor, storage_cache, version_state, memory_cache
                )
                all_chunk_engines[tensor] = storage_chunk_engine
                break
            except (JSONDecodeError, KeyError):
                if i == num_tries - 1:
                    raise
    return all_chunk_engines


def add_cache_to_dataset_slice(
    dataset_slice: hub.Dataset,
    tensors: List[str],
) -> hub.Dataset:
    base_storage = get_base_storage(dataset_slice.storage)
    # 64 to account for potentially big encoder corresponding to each tensor
    # TODO: adjust this size once we get rid of cachable
    cache_size = 64 * len(tensors) * MB
    cached_store = LRUCache(MemoryProvider(), base_storage, cache_size)
    dataset_slice = hub.Dataset(
        cached_store,
        index=dataset_slice.index,
        group_index=dataset_slice.group_index,  # type: ignore
        read_only=dataset_slice.read_only,
        verbose=False,
    )
    return dataset_slice


def check_transform_data_in(data_in, scheduler: str) -> None:
    """Checks whether the data_in for a transform is valid or not."""
    if not hasattr(data_in, "__getitem__"):
        raise InvalidInputDataError("__getitem__")
    if not hasattr(data_in, "__len__"):
        raise InvalidInputDataError("__len__")
    if isinstance(data_in, hub.Dataset):
        input_base_storage = get_base_storage(data_in.storage)
        if isinstance(input_base_storage, MemoryProvider) and scheduler not in [
            "serial",
            "threaded",
        ]:
            raise InvalidInputDataError(
                f"Transforms with data_in as a Dataset having base storage as MemoryProvider are only supported in threaded and serial mode. Current mode is {scheduler}."
            )


def check_transform_ds_out(ds_out: hub.Dataset, scheduler: str) -> None:
    """Checks whether the ds_out for a transform is valid or not."""
    if ds_out._read_only:
        raise InvalidOutputDatasetError
    tensors = list(ds_out.tensors)
    for tensor in tensors:
        if len(ds_out[tensor]) != len(ds_out):
            raise InvalidOutputDatasetError(
                "One or more tensors of the ds_out have different lengths. Transform only supports ds_out having same number of samples for each tensor (This includes empty datasets that have 0 samples per tensor)."
            )

    output_base_storage = get_base_storage(ds_out.storage)
    if isinstance(output_base_storage, MemoryProvider) and scheduler not in [
        "serial",
        "threaded",
    ]:
        raise InvalidOutputDatasetError(
            f"Transforms with ds_out having base storage as MemoryProvider are only supported in threaded and serial mode. Current mode is {scheduler}."
        )


def get_pbar_description(transform_functions: List):
    """Returns the description string for a hub.compute evaluation progress bar. Incoming list should be a list of `TransformFunction`s."""

    num_funcs = len(transform_functions)
    if num_funcs == 0:
        return "Evaluating"

    func_names: List[str] = []
    for transform_function in transform_functions:
        func_names.append(transform_function.func.__name__)

    if num_funcs == 1:
        return f"Evaluating {func_names[0]}"

    names_desc = ", ".join(func_names)
    return f"Evaluating [{names_desc}]"



from typing import Set
from hub.core.storage.cachable import Cachable


class CommitChunkSet(Cachable):
    """Stores set of chunks stored for a particular tensor in a commit."""

    def __init__(self) -> None:
        self.chunks: Set[str] = set()

    def tobytes(self) -> bytes:
        """Dumps self.chunks in csv format."""
        return bytes(",".join(self.chunks), "utf-8")

    @classmethod
    def frombuffer(cls, buffer: bytes):
        """Loads a CommitChunkSet from a buffer."""
        instance = cls()
        instance.chunks = set(buffer.decode("utf-8").split(","))
        return instance

    @property
    def nbytes(self) -> int:
        if not self.chunks:
            return 0
        return 8 + ((len(self.chunks) - 1) * 9)

    def add(self, chunk_name: str) -> None:
        """Adds a new chunk name to the CommitChunkSet."""
        self.chunks.add(chunk_name)



import hub
import pytest
import numpy as np
from click.testing import CliRunner
from hub.core.storage.memory import MemoryProvider
from hub.util.remove_cache import remove_memory_cache
from hub.util.check_installation import ray_installed
from hub.util.exceptions import InvalidOutputDatasetError, TransformError
from hub.tests.common import parametrize_num_workers
from hub.tests.dataset_fixtures import enabled_datasets, enabled_non_gcs_datasets
from hub.util.transform import get_pbar_description
import sys
import hub


# TODO progressbar is disabled while running tests on mac for now
if sys.platform == "darwin":
    defs = hub.core.transform.transform.Pipeline.eval.__defaults__
    defs = defs[:-1] + (False,)
    hub.core.transform.transform.Pipeline.eval.__defaults__ = defs

    defs = hub.core.transform.transform.TransformFunction.eval.__defaults__
    defs = defs[:-1] + (False,)
    hub.core.transform.transform.TransformFunction.eval.__defaults__ = defs


# github actions can only support 2 workers
TRANSFORM_TEST_NUM_WORKERS = 2

# github actions can only support 2 workers
TRANSFORM_TEST_NUM_WORKERS = 2

all_compressions = pytest.mark.parametrize("sample_compression", [None, "png", "jpeg"])

schedulers = ["threaded", "processed"]
schedulers = schedulers + ["ray"] if ray_installed() else schedulers
all_schedulers = pytest.mark.parametrize("scheduler", schedulers)


@hub.compute
def fn1(sample_in, samples_out, mul=1, copy=1):
    for _ in range(copy):
        samples_out.image.append(np.ones((337, 200)) * sample_in * mul)
        samples_out.label.append(np.ones((1,)) * sample_in * mul)


@hub.compute
def fn2(sample_in, samples_out, mul=1, copy=1):
    for _ in range(copy):
        samples_out.image.append(sample_in.image.numpy() * mul)
        samples_out.label.append(sample_in.label.numpy() * mul)


@hub.compute
def fn3(sample_in, samples_out, mul=1, copy=1):
    for _ in range(copy):
        samples_out.image.append(np.ones((1310, 2087)) * sample_in * mul)
        samples_out.label.append(np.ones((13,)) * sample_in * mul)


@hub.compute
def read_image(sample_in, samples_out):
    samples_out.image.append(hub.read(sample_in))


@hub.compute
def crop_image(sample_in, samples_out, copy=1):
    for _ in range(copy):
        samples_out.image.append(sample_in.image.numpy()[:100, :100, :])


@hub.compute
def filter_tr(sample_in, sample_out):
    if sample_in % 2 == 0:
        sample_out.image.append(sample_in * np.ones((100, 100)))


@all_schedulers
@enabled_non_gcs_datasets
def test_single_transform_hub_dataset(ds, scheduler):
    data_in = hub.dataset("./test/single_transform_hub_dataset", overwrite=True)
    with data_in:
        data_in.create_tensor("image")
        data_in.create_tensor("label")
        for i in range(1, 100):
            data_in.image.append(i * np.ones((i, i)))
            data_in.label.append(i * np.ones((1,)))
    ds_out = ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    if (
        isinstance(remove_memory_cache(ds.storage), MemoryProvider)
        and scheduler != "threaded"
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        with pytest.raises(InvalidOutputDatasetError):
            fn2(copy=1, mul=2).eval(
                data_in,
                ds_out,
                num_workers=TRANSFORM_TEST_NUM_WORKERS,
                scheduler=scheduler,
            )
        data_in.delete()
        return

    fn2(copy=1, mul=2).eval(
        data_in, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, scheduler=scheduler
    )
    assert len(ds_out) == 99
    for index in range(1, 100):
        np.testing.assert_array_equal(
            ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
        )
        np.testing.assert_array_equal(
            ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
        )

    assert ds_out.image.shape_interval.lower == (99, 1, 1)
    assert ds_out.image.shape_interval.upper == (99, 99, 99)
    data_in.delete()


@enabled_datasets
def test_groups(ds):
    with CliRunner().isolated_filesystem():
        with hub.dataset("./test/transform_hub_in_generic") as data_in:
            data_in.create_tensor("data/image")
            data_in.create_tensor("data/label")
            for i in range(1, 100):
                data_in.data.image.append(i * np.ones((i, i)))
                data_in.data.label.append(i * np.ones((1,)))
        data_in = hub.dataset("./test/transform_hub_in_generic")
        ds_out = ds
        ds_out.create_tensor("stuff/image")
        ds_out.create_tensor("stuff/label")

        data_in = data_in.data
        ds_out = ds_out.stuff

        fn2(copy=1, mul=2).eval(data_in, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS)
        assert len(ds_out) == 99
        for index in range(1, 100):
            np.testing.assert_array_equal(
                ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
            )
            np.testing.assert_array_equal(
                ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
            )

        assert ds_out.image.shape_interval.lower == (99, 1, 1)
        assert ds_out.image.shape_interval.upper == (99, 99, 99)


@enabled_non_gcs_datasets
@parametrize_num_workers
@all_schedulers
def test_single_transform_hub_dataset_htypes(ds, num_workers, scheduler):
    data_in = hub.dataset("./test/single_transform_hub_dataset_htypes", overwrite=True)
    with data_in:
        data_in.create_tensor("image", htype="image", sample_compression="png")
        data_in.create_tensor("label", htype="class_label")
        for i in range(1, 100):
            data_in.image.append(i * np.ones((i, i), dtype="uint8"))
            data_in.label.append(i * np.ones((1,), dtype="uint32"))
    ds_out = ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    if (
        isinstance(remove_memory_cache(ds.storage), MemoryProvider)
        and scheduler != "threaded"
        and num_workers > 0
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        # num_workers = 0 automatically does single threaded irrespective of the scheduler
        with pytest.raises(InvalidOutputDatasetError):
            fn2(copy=1, mul=2).eval(
                data_in, ds_out, num_workers=num_workers, scheduler=scheduler
            )
        data_in.delete()
        return
    fn2(copy=1, mul=2).eval(
        data_in, ds_out, num_workers=num_workers, scheduler=scheduler
    )
    assert len(ds_out) == 99
    for index in range(1, 100):
        np.testing.assert_array_equal(
            ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
        )
        np.testing.assert_array_equal(
            ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
        )

    assert ds_out.image.shape_interval.lower == (99, 1, 1)
    assert ds_out.image.shape_interval.upper == (99, 99, 99)
    data_in.delete()


@all_schedulers
@enabled_non_gcs_datasets
def test_chain_transform_list_small(ds, scheduler):
    ls = [i for i in range(100)]
    ds_out = ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    pipeline = hub.compose([fn1(mul=5, copy=2), fn2(mul=3, copy=3)])
    if (
        isinstance(remove_memory_cache(ds.storage), MemoryProvider)
        and scheduler != "threaded"
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        with pytest.raises(InvalidOutputDatasetError):
            pipeline.eval(
                ls, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, scheduler=scheduler
            )
        return
    pipeline.eval(
        ls, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, scheduler=scheduler
    )
    assert len(ds_out) == 600
    for i in range(100):
        for index in range(6 * i, 6 * i + 6):
            np.testing.assert_array_equal(
                ds_out[index].image.numpy(), 15 * i * np.ones((337, 200))
            )
            np.testing.assert_array_equal(
                ds_out[index].label.numpy(), 15 * i * np.ones((1,))
            )


@all_schedulers
@enabled_non_gcs_datasets
@pytest.mark.xfail(raises=TransformError, strict=False)
def test_chain_transform_list_big(ds, scheduler):
    ls = [i for i in range(2)]
    ds_out = ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    pipeline = hub.compose([fn3(mul=5, copy=2), fn2(mul=3, copy=3)])
    if (
        isinstance(remove_memory_cache(ds.storage), MemoryProvider)
        and scheduler != "threaded"
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        with pytest.raises(InvalidOutputDatasetError):
            pipeline.eval(
                ls, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, scheduler=scheduler
            )
        return
    pipeline.eval(
        ls, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, scheduler=scheduler
    )
    assert len(ds_out) == 8
    for i in range(2):
        for index in range(4 * i, 4 * i + 4):
            np.testing.assert_array_equal(
                ds_out[index].image.numpy(), 15 * i * np.ones((1310, 2087))
            )
            np.testing.assert_array_equal(
                ds_out[index].label.numpy(), 15 * i * np.ones((13,))
            )


@all_schedulers
@all_compressions
@enabled_non_gcs_datasets
def test_transform_hub_read(ds, cat_path, sample_compression, scheduler):
    data_in = [cat_path] * 10
    ds_out = ds
    ds_out.create_tensor("image", htype="image", sample_compression=sample_compression)

    if (
        isinstance(remove_memory_cache(ds.storage), MemoryProvider)
        and scheduler != "threaded"
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        with pytest.raises(InvalidOutputDatasetError):
            read_image().eval(
                data_in,
                ds_out,
                num_workers=TRANSFORM_TEST_NUM_WORKERS,
                scheduler=scheduler,
            )
        return

    read_image().eval(
        data_in, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, scheduler=scheduler
    )
    assert len(ds_out) == 10
    for i in range(10):
        assert ds_out.image[i].numpy().shape == (900, 900, 3)
        np.testing.assert_array_equal(ds_out.image[i].numpy(), ds_out.image[0].numpy())


@all_schedulers
@all_compressions
@enabled_non_gcs_datasets
def test_transform_hub_read_pipeline(ds, cat_path, sample_compression, scheduler):
    data_in = [cat_path] * 10
    ds_out = ds
    ds_out.create_tensor("image", htype="image", sample_compression=sample_compression)
    pipeline = hub.compose([read_image(), crop_image(copy=2)])
    if (
        isinstance(remove_memory_cache(ds.storage), MemoryProvider)
        and scheduler != "threaded"
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        with pytest.raises(InvalidOutputDatasetError):
            pipeline.eval(
                data_in,
                ds_out,
                num_workers=TRANSFORM_TEST_NUM_WORKERS,
                scheduler=scheduler,
            )
        return
    pipeline.eval(
        data_in, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, scheduler=scheduler
    )
    assert len(ds_out) == 20
    for i in range(20):
        assert ds_out.image[i].numpy().shape == (100, 100, 3)
        np.testing.assert_array_equal(ds_out.image[i].numpy(), ds_out.image[0].numpy())


@enabled_non_gcs_datasets
def test_hub_like(ds, scheduler="threaded"):
    with CliRunner().isolated_filesystem():
        data_in = ds
        with data_in:
            data_in.create_tensor("image", htype="image", sample_compression="png")
            data_in.create_tensor("label", htype="class_label")
            for i in range(1, 100):
                data_in.image.append(i * np.ones((i, i), dtype="uint8"))
                data_in.label.append(i * np.ones((1,), dtype="uint32"))
        ds_out = hub.like("./transform_hub_like", data_in)
        if (
            isinstance(remove_memory_cache(ds.storage), MemoryProvider)
            and scheduler != "threaded"
        ):
            # any scheduler other than `threaded` will not work with a dataset stored in memory
            with pytest.raises(InvalidOutputDatasetError):
                fn2(copy=1, mul=2).eval(
                    data_in,
                    ds_out,
                    num_workers=TRANSFORM_TEST_NUM_WORKERS,
                    scheduler=scheduler,
                )
            return
        fn2(copy=1, mul=2).eval(
            data_in, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, scheduler=scheduler
        )
        assert len(ds_out) == 99
        for index in range(1, 100):
            np.testing.assert_array_equal(
                ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
            )
            np.testing.assert_array_equal(
                ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
            )

        assert ds_out.image.shape_interval.lower == (99, 1, 1)
        assert ds_out.image.shape_interval.upper == (99, 99, 99)


def test_transform_empty(local_ds):
    local_ds.create_tensor("image")

    ls = list(range(10))
    filter_tr().eval(ls, local_ds)

    assert len(local_ds) == 5

    for i in range(5):
        np.testing.assert_array_equal(
            local_ds[i].image.numpy(), 2 * i * np.ones((100, 100))
        )


def test_pbar_description():
    assert get_pbar_description([fn1()]) == "Evaluating fn1"
    assert get_pbar_description([fn1(), fn2()]) == "Evaluating [fn1, fn2]"
    assert get_pbar_description([fn1(), fn1()]) == "Evaluating [fn1, fn1]"
    assert (
        get_pbar_description([fn1(), fn1(), read_image()])
        == "Evaluating [fn1, fn1, read_image]"
    )


def test_bad_transform(memory_ds):
    ds = memory_ds
    ds.create_tensor("x")
    ds.create_tensor("y")
    with ds:
        ds.x.extend(np.random.rand(10, 1))
        ds.y.extend(np.random.rand(10, 1))
    ds2 = hub.like("mem://dummy2", ds)

    @hub.compute
    def fn_filter(sample_in, sample_out):
        sample_out.y.append(sample_in.y.numpy())
        return sample_out

    with pytest.raises(TransformError):
        fn_filter().eval(ds, ds2, progressbar=True)


def test_transform_persistance(local_ds_generator, num_workers=2, scheduler="threaded"):
    data_in = hub.dataset("./test/single_transform_hub_dataset_htypes", overwrite=True)
    with data_in:
        data_in.create_tensor("image", htype="image", sample_compression="png")
        data_in.create_tensor("label", htype="class_label")
        for i in range(1, 100):
            data_in.image.append(i * np.ones((i, i), dtype="uint8"))
            data_in.label.append(i * np.ones((1,), dtype="uint32"))
    ds_out = local_ds_generator()
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    if (
        isinstance(remove_memory_cache(ds_out.storage), MemoryProvider)
        and scheduler != "threaded"
        and num_workers > 0
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        # num_workers = 0 automatically does single threaded irrespective of the scheduler
        with pytest.raises(InvalidOutputDatasetError):
            fn2(copy=1, mul=2).eval(
                data_in, ds_out, num_workers=num_workers, scheduler=scheduler
            )
        data_in.delete()
        return
    fn2(copy=1, mul=2).eval(
        data_in, ds_out, num_workers=num_workers, scheduler=scheduler
    )

    def test_ds_out():
        assert len(ds_out) == 99
        for index in range(1, 100):
            np.testing.assert_array_equal(
                ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
            )
            np.testing.assert_array_equal(
                ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
            )

        assert ds_out.image.shape_interval.lower == (99, 1, 1)
        assert ds_out.image.shape_interval.upper == (99, 99, 99)

    test_ds_out()
    ds_out = local_ds_generator()
    test_ds_out()

    data_in.delete()



from hub.api.dataset import Dataset
from hub.util.exceptions import (
    KaggleDatasetAlreadyDownloadedError,
    SamePathException,
    KaggleMissingCredentialsError,
    ExternalCommandError,
)
from click.testing import CliRunner
from hub.tests.common import get_dummy_data_path
import pytest
import os
import hub


def test_ingestion_simple(local_ds: Dataset, hub_kaggle_credentials):
    with CliRunner().isolated_filesystem():
        kaggle_path = os.path.join(local_ds.path, "unstructured_kaggle_data_simple")
        username, key = hub_kaggle_credentials

        ds = hub.ingest_kaggle(
            tag="andradaolteanu/birdcall-recognition-data",
            src=kaggle_path,
            dest=local_ds.path,
            images_compression="jpeg",
            kaggle_credentials={"username": username, "key": key},
            progress_bar=False,
            summary=False,
            overwrite=False,
        )

        assert list(ds.tensors.keys()) == ["images", "labels"]
        assert ds["labels"].numpy().shape == (10, 1)


def test_ingestion_sets(local_ds: Dataset, hub_kaggle_credentials):
    with CliRunner().isolated_filesystem():
        kaggle_path = os.path.join(local_ds.path, "unstructured_kaggle_data_sets")
        username, key = hub_kaggle_credentials

        ds = hub.ingest_kaggle(
            tag="thisiseshan/bird-classes",
            src=kaggle_path,
            dest=local_ds.path,
            images_compression="jpeg",
            kaggle_credentials={"username": username, "key": key},
            progress_bar=False,
            summary=False,
            overwrite=False,
        )

        assert list(ds.tensors.keys()) == [
            "test/images",
            "test/labels",
            "train/images",
            "train/labels",
        ]
        assert ds["test/images"].numpy().shape == (3, 200, 200, 3)
        assert ds["test/labels"].numpy().shape == (3, 1)
        assert ds["test/labels"].info.class_names == ("class0", "class1", "class2")

        assert ds["train/images"].numpy().shape == (3, 200, 200, 3)
        assert ds["train/labels"].numpy().shape == (3, 1)
        assert ds["train/labels"].info.class_names == ("class0", "class1", "class2")


def test_kaggle_exception(local_ds: Dataset, hub_kaggle_credentials):
    with CliRunner().isolated_filesystem():
        kaggle_path = os.path.join(local_ds.path, "unstructured_kaggle_data")
        dummy_path = get_dummy_data_path("tests_auto/image_classification")
        username, key = hub_kaggle_credentials

        with pytest.raises(SamePathException):
            hub.ingest_kaggle(
                tag="thisiseshan/bird-classes",
                src=dummy_path,
                dest=dummy_path,
                images_compression="jpeg",
                kaggle_credentials={"username": username, "key": key},
                progress_bar=False,
                summary=False,
                overwrite=False,
            )

        with pytest.raises(KaggleMissingCredentialsError):
            hub.ingest_kaggle(
                tag="thisiseshan/bird-classes",
                src=kaggle_path,
                dest=local_ds.path,
                images_compression="jpeg",
                kaggle_credentials={"not_username": "not_username"},
                progress_bar=False,
                summary=False,
                overwrite=False,
            )

        with pytest.raises(KaggleMissingCredentialsError):
            hub.ingest_kaggle(
                tag="thisiseshan/bird-classes",
                src=kaggle_path,
                dest=local_ds.path,
                images_compression="jpeg",
                kaggle_credentials={"username": "thisiseshan", "not_key": "not_key"},
                progress_bar=False,
                summary=False,
                overwrite=False,
            )

        with pytest.raises(ExternalCommandError):
            hub.ingest_kaggle(
                tag="thisiseshan/invalid-dataset",
                src=kaggle_path,
                dest=local_ds.path,
                images_compression="jpeg",
                kaggle_credentials={"username": username, "key": key},
                progress_bar=False,
                summary=False,
                overwrite=False,
            )

        hub.ingest_kaggle(
            tag="thisiseshan/bird-classes",
            src=kaggle_path,
            dest=local_ds.path,
            images_compression="jpeg",
            kaggle_credentials={"username": username, "key": key},
            progress_bar=False,
            summary=False,
            overwrite=False,
        )

        with pytest.raises(KaggleDatasetAlreadyDownloadedError):
            hub.ingest_kaggle(
                tag="thisiseshan/bird-classes",
                src=kaggle_path,
                dest=local_ds.path,
                images_compression="jpeg",
                kaggle_credentials={"username": username, "key": key},
                progress_bar=False,
                summary=False,
                overwrite=False,
            )



import os
import pickle
import warnings
import hub
from typing import Callable, Union, Optional, Dict, Tuple, Sequence
from hub.core.storage import MemoryProvider, LRUCache
from hub.util.dataset import try_flushing
from hub.util.remove_cache import get_base_storage
from hub.util.iterable_ordered_dict import IterableOrderedDict
from hub.util.exceptions import (
    DatasetUnsupportedPytorch,
    ModuleNotInstalledException,
    TensorDoesNotExistError,
    SampleDecompressionError,
)
from hub.constants import MB
from .common import convert_fn as default_convert_fn, collate_fn as default_collate_fn


def dataset_to_pytorch(
    dataset,
    transform: Optional[Callable] = None,
    tensors: Optional[Sequence[str]] = None,
    num_workers: int = 1,
    batch_size: Optional[int] = 1,
    drop_last: bool = False,
    collate_fn: Optional[Callable] = None,
    pin_memory: bool = False,
    shuffle: bool = False,
    buffer_size: int = 10 * 1000,
    use_local_cache: bool = False,
    python_version_warning: bool = True,
):
    try_flushing(dataset)

    global torch
    try:
        import torch
    except ModuleNotFoundError:
        raise ModuleNotInstalledException(
            "'torch' should be installed to convert the Dataset into pytorch format"
        )

    pytorch_ds = TorchDataset(
        dataset,
        transform,
        tensors,
        shuffle,
        python_version_warning=python_version_warning,
    )

    if collate_fn is None:
        collate_fn = default_convert_fn if batch_size is None else default_collate_fn

    return torch.utils.data.DataLoader(  # type: ignore
        pytorch_ds,
        num_workers=num_workers,
        batch_size=batch_size,
        drop_last=drop_last,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
        shuffle=shuffle,
    )


class TorchDataset:
    def __init__(
        self,
        dataset,
        transform: Optional[Callable] = None,
        tensors: Optional[Sequence[str]] = None,
        shuffle: bool = False,
        python_version_warning: bool = True,
    ):

        if python_version_warning:
            warning_message = ""
            if os.name == "nt":
                warning_message += f"Windows OS detected. Pytorch iteration speeds are up to 500% faster using linux/macOS along with Python version >= 3.8. "
            else:
                warning_message += f"Python version < 3.8 detected. Pytorch iteration speeds are up to 500% faster on Python version >= 3.8.\n"

            warning_message += "This version will also not utilize the buffer_size and use_local_cache arguments.\n"
            if shuffle:
                warning_message += "Pytorch iteration with shuffling will also be very slow. Use linux/macOS with python >= 3.8 to speed it up.\n"
            warnings.warn(warning_message)

        self.dataset = None
        self._worker_range = None
        base_storage = get_base_storage(dataset.storage)
        if isinstance(base_storage, MemoryProvider):
            raise DatasetUnsupportedPytorch(
                "Datasets whose underlying storage is MemoryProvider are not supported for Pytorch iteration."
            )
        self.pickled_storage = pickle.dumps(base_storage)
        self.index = dataset.index
        self.group_index = dataset.group_index
        self.length = len(dataset)
        self.transform = transform
        if tensors is None:
            self.tensor_keys = list(dataset.tensors)
        else:
            for t in tensors:
                if t not in dataset.tensors:
                    raise TensorDoesNotExistError(t)
            self.tensor_keys = list(tensors)
        self._num_bad_samples = 0

    def _apply_transform(self, sample: Union[Dict, Tuple]):
        return self.transform(sample) if self.transform else sample

    def _init_ds(self):
        """
        For each process, dataset should be independently loaded
        """
        if self.dataset is None:
            storage = pickle.loads(self.pickled_storage)

            # creating a new cache for each process
            cache_size = 32 * MB * len(self.tensor_keys)
            cached_storage = LRUCache(MemoryProvider(), storage, cache_size)
            self.dataset = hub.Dataset(
                storage=cached_storage,
                index=self.index,
                group_index=self.group_index,
                verbose=False,
            )

    def __len__(self):
        return self.length

    def _get(self, index: int):
        self._init_ds()
        sample = IterableOrderedDict()
        # pytorch doesn't support certain dtypes, which are type casted to another dtype below
        for key in self.tensor_keys:
            try:
                item = self.dataset[key][index].numpy()  # type: ignore
            except SampleDecompressionError as e:
                warnings.warn(
                    f"Skipping corrupt {self.dataset[key].meta.sample_compression} sample."  # type: ignore
                )
                return None
            if item.dtype == "uint16":
                item = item.astype("int32")
            elif item.dtype in ["uint32", "uint64"]:
                item = item.astype("int64")
            sample[key] = item

        return self._apply_transform(sample)

    def __getitem__(self, index: int):
        while True:
            next_good_sample_index = index + self._num_bad_samples
            if next_good_sample_index >= self.length:
                raise StopIteration()  # DataLoader expects StopIteration, not IndexError
            val = self._get(next_good_sample_index)
            if val is None:
                self._num_bad_samples += 1
            else:
                return val

    def __iter__(self):
        for index in range(len(self)):
            val = self[index]
            if val is not None:
                yield val



import random
import time
import hashlib
import pickle
from typing import Any, Dict
from hub.client.log import logger
from hub.constants import FIRST_COMMIT_ID
from hub.core.fast_forwarding import ffw_dataset_meta
from hub.core.meta.dataset_meta import DatasetMeta
from hub.core.storage.cachable import Cachable
from hub.core.version_control.commit_node import CommitNode  # type: ignore
from hub.core.version_control.commit_chunk_set import CommitChunkSet  # type: ignore
from hub.core.storage import LRUCache
from hub.util.exceptions import CallbackInitializationError, CheckoutError
from hub.util.keys import (
    get_chunk_id_encoder_key,
    get_dataset_info_key,
    get_dataset_meta_key,
    get_tensor_info_key,
    get_tensor_meta_key,
    get_tensor_commit_chunk_set_key,
    get_version_control_info_key,
)


def generate_hash() -> str:
    hsh = hashlib.sha1()
    hsh.update(str(time.time()).encode("utf-8"))
    hsh.update(random.randrange(0, 1000000).to_bytes(4, "big"))
    return hsh.hexdigest()


def commit(
    version_state: Dict[str, Any], storage: LRUCache, message: str = None
) -> None:
    """Modifies the version state to reflect the commit and also copies required data to the new commit directory."""
    storage.check_readonly()
    # if not the head node, checkout to an auto branch that is newly created
    auto_checkout(version_state, storage)
    stored_commit_id = version_state["commit_id"]
    version_state["commit_id"] = generate_hash()
    new_node = CommitNode(version_state["branch"], version_state["commit_id"])
    version_state["commit_node"].add_successor(new_node, message)
    version_state["commit_node"] = new_node
    version_state["branch_commit_map"][version_state["branch"]] = version_state[
        "commit_id"
    ]
    version_state["commit_node_map"][version_state["commit_id"]] = new_node
    save_version_info(version_state, storage)
    copy_metas(
        stored_commit_id,
        version_state["commit_id"],
        storage,
        version_state["full_tensors"],
    )
    load_meta(storage, version_state)


def checkout(
    version_state: Dict[str, Any],
    storage: LRUCache,
    address: str,
    create: bool = False,
) -> None:
    """Modifies the version state to reflect the checkout and also copies required data to the new branch directory if a new one is being created."""
    original_commit_id = version_state["commit_id"]

    if address in version_state["branch_commit_map"].keys():
        if create:
            raise CheckoutError(f"Can't create new branch, '{address}' already exists.")
        version_state["branch"] = address
        new_commit_id = version_state["branch_commit_map"][address]
        if original_commit_id == new_commit_id:
            return
        version_state["commit_id"] = new_commit_id
        version_state["commit_node"] = version_state["commit_node_map"][new_commit_id]
    elif address in version_state["commit_node_map"].keys():
        if create:
            raise CheckoutError(
                f"Can't create new branch, commit '{address}' already exists."
            )
        if address == original_commit_id:
            return
        version_state["commit_id"] = address
        version_state["commit_node"] = version_state["commit_node_map"][address]
        version_state["branch"] = version_state["commit_node"].branch
    elif create:
        storage.check_readonly()
        # if the original commit is head of the branch and has data, auto commit and checkout to original commit before creating new branch
        auto_commit(version_state, storage, address)
        new_commit_id = generate_hash()
        new_node = CommitNode(address, new_commit_id)
        version_state["commit_node"].add_child(new_node)
        version_state["commit_id"] = new_commit_id
        version_state["commit_node"] = new_node
        version_state["branch"] = address
        version_state["commit_node_map"][new_commit_id] = new_node
        version_state["branch_commit_map"][address] = new_commit_id
        save_version_info(version_state, storage)
        copy_metas(
            original_commit_id,
            new_commit_id,
            storage,
            version_state["full_tensors"],
        )
    else:
        raise CheckoutError(
            f"Address {address} not found. If you want to create a new branch, use checkout with create=True"
        )

    discard_old_metas(
        original_commit_id,
        storage,
        version_state["full_tensors"],
    )
    load_meta(storage, version_state)


def copy_metas(
    src_commit_id: str, dest_commit_id: str, storage: LRUCache, tensors: Dict
) -> None:
    """Copies meta data from one commit to another."""

    src_dataset_meta_key = get_dataset_meta_key(src_commit_id)
    dest_dataset_meta_key = get_dataset_meta_key(dest_commit_id)
    src_dataset_meta = storage[src_dataset_meta_key]
    if isinstance(src_dataset_meta, Cachable):
        storage[dest_dataset_meta_key] = src_dataset_meta.copy()
    else:
        storage[dest_dataset_meta_key] = src_dataset_meta

    try:
        src_dataset_info_key = get_dataset_info_key(src_commit_id)
        dest_dataset_info_key = get_dataset_info_key(dest_commit_id)
        src_dataset_info = storage[src_dataset_info_key]
        if isinstance(src_dataset_info, Cachable):
            storage[dest_dataset_info_key] = src_dataset_info.copy()
        else:
            storage[dest_dataset_info_key] = src_dataset_info
    except (KeyError, CallbackInitializationError):
        pass

    tensor_list = list(tensors.keys())

    for tensor in tensor_list:
        src_tensor_meta_key = get_tensor_meta_key(tensor, src_commit_id)
        dest_tensor_meta_key = get_tensor_meta_key(tensor, dest_commit_id)
        src_tensor_meta = storage[src_tensor_meta_key]
        if isinstance(src_tensor_meta, Cachable):
            storage[dest_tensor_meta_key] = src_tensor_meta.copy()
        else:
            storage[dest_tensor_meta_key] = src_tensor_meta

        try:
            src_chunk_id_encoder_key = get_chunk_id_encoder_key(tensor, src_commit_id)
            dest_chunk_id_encoder_key = get_chunk_id_encoder_key(tensor, dest_commit_id)
            src_chunk_id_encoder = storage[src_chunk_id_encoder_key]
            if isinstance(src_chunk_id_encoder, Cachable):
                storage[dest_chunk_id_encoder_key] = src_chunk_id_encoder.copy()
            else:
                storage[dest_chunk_id_encoder_key] = src_chunk_id_encoder
        except (KeyError, CallbackInitializationError):
            pass

        try:
            src_tensor_info_key = get_tensor_info_key(tensor, src_commit_id)
            dest_tensor_info_key = get_tensor_info_key(tensor, dest_commit_id)
            src_tensor_info = storage[src_tensor_info_key]
            if isinstance(src_tensor_info, Cachable):
                storage[dest_tensor_info_key] = src_tensor_info.copy()
            else:
                storage[dest_tensor_info_key] = src_tensor_info
        except (KeyError, CallbackInitializationError):
            pass

    storage.flush()


def discard_old_metas(
    src_commit_id: str,
    storage: LRUCache,
    tensors: Dict,
):
    """Discards the metas of the previous commit from cache, during checkout or when a new commit is made."""
    all_src_keys = []
    src_dataset_meta_key = get_dataset_meta_key(src_commit_id)
    all_src_keys.append(src_dataset_meta_key)

    src_dataset_info_key = get_dataset_info_key(src_commit_id)
    all_src_keys.append(src_dataset_info_key)

    tensor_list = list(tensors.keys())

    for tensor in tensor_list:
        src_tensor_meta_key = get_tensor_meta_key(tensor, src_commit_id)
        all_src_keys.append(src_tensor_meta_key)

        src_chunk_id_encoder_key = get_chunk_id_encoder_key(tensor, src_commit_id)
        all_src_keys.append(src_chunk_id_encoder_key)

        src_tensor_info_key = get_tensor_info_key(tensor, src_commit_id)
        all_src_keys.append(src_tensor_info_key)

    for key in all_src_keys:
        storage.dirty_keys.discard(key)
        if key in storage.lru_sizes:
            size = storage.lru_sizes.pop(key)
            storage.cache_used -= size
        try:
            del storage.cache_storage[key]
        except (KeyError, CallbackInitializationError):
            pass


def save_version_info(version_state: Dict[str, Any], storage: LRUCache) -> None:
    """Saves the current version info to the storage."""
    version_info = {
        "commit_node_map": version_state["commit_node_map"],
        "branch_commit_map": version_state["branch_commit_map"],
    }
    storage[get_version_control_info_key()] = pickle.dumps(version_info)


def auto_checkout(version_state: Dict[str, Any], storage: LRUCache) -> None:
    """Automatically checks out if current node is not the head node of the branch. This may happen either during commit/setitem/append/extend/create_tensor/info updates."""
    if version_state["commit_node"].commit_time is not None:
        current_branch = version_state["branch"]
        auto_branch = f"auto_{generate_hash()}"
        logger.info(
            f"Automatically checking out to branch '{auto_branch}' as not currently at the head node of branch '{current_branch}'."
        )
        checkout(version_state, storage, auto_branch, True)


def auto_commit(version_state: Dict[str, Any], storage: LRUCache, address: str) -> None:
    """Automatically commits to the current branch before a checkout to a newly created branch if the current node is the head node and has uncommitted data."""
    commit_node = version_state["commit_node"]
    if not commit_node.commit_time and commit_has_data(version_state, storage):

        original_commit_id = version_state["commit_id"]
        branch = version_state["branch"]
        logger.info(
            f"Auto commiting to branch '{branch}' as currently at head node with uncommitted changes."
        )
        commit(
            version_state,
            storage,
            f"auto commit before checkout to {address}",
        )
        checkout(version_state, storage, original_commit_id, False)


def commit_has_data(version_state: Dict[str, Any], storage: LRUCache) -> bool:
    """Checks if the current commit has any data present in it or not."""
    commit_id = version_state["commit_id"]
    for tensor in version_state["full_tensors"].keys():
        if commit_id == FIRST_COMMIT_ID:
            # if the first commit has even a single tensor i.e. it entered the for loop, it has data
            return True
        key = get_tensor_commit_chunk_set_key(tensor, commit_id)
        if commit_chunk_set_exists(version_state, storage, tensor):
            enc = storage.get_cachable(key, CommitChunkSet)
            if enc.chunks:
                return True
    return False


def commit_chunk_set_exists(
    version_state: Dict[str, Any], storage: LRUCache, tensor: str
) -> bool:
    """Checks if the commit chunk set exists for the given tensor in the current commit."""
    try:
        commit_id = version_state["commit_id"]
        key = get_tensor_commit_chunk_set_key(tensor, commit_id)
        storage[key]
        return True
    except KeyError:
        return False


def load_meta(storage, version_state):
    """Loads the meta info for the version state."""
    from hub.core.tensor import Tensor

    meta_key = get_dataset_meta_key(version_state["commit_id"])
    meta = storage.get_cachable(meta_key, DatasetMeta)
    ffw_dataset_meta(meta)
    version_state["meta"] = meta
    _tensors = version_state["full_tensors"]
    _tensors.clear()

    for tensor_name in meta.tensors:
        _tensors[tensor_name] = Tensor(tensor_name, storage, version_state)



from hub.compression import get_compression_type
from hub.util.exceptions import (
    SampleCompressionError,
    TensorMetaMissingRequiredValue,
    TensorMetaMutuallyExclusiveKeysError,
    UnsupportedCompressionError,
)
import pytest
from hub.core.tensor import Tensor
from hub.tests.common import TENSOR_KEY, assert_images_close
from hub.tests.dataset_fixtures import enabled_datasets
import numpy as np

import hub
from hub.core.dataset import Dataset
from miniaudio import mp3_read_file_f32, flac_read_file_f32, wav_read_file_f32  # type: ignore


def _populate_compressed_samples(tensor: Tensor, cat_path, flower_path, count=1):
    for _ in range(count):
        tensor.append(hub.read(cat_path))
        tensor.append(hub.read(flower_path))
        tensor.append(np.ones((100, 100, 4), dtype="uint8"))
        tensor.append(
            np.ones((100, 100, 4), dtype=int).tolist()
        )  # test safe downcasting of python scalars

        tensor.extend(
            [
                hub.read(flower_path),
                hub.read(cat_path),
            ]
        )


@enabled_datasets
def test_populate_compressed_samples(ds: Dataset, cat_path, flower_path):
    images = ds.create_tensor(TENSOR_KEY, htype="image", sample_compression="png")

    assert images.meta.dtype == "uint8"
    assert images.meta.sample_compression == "png"

    _populate_compressed_samples(images, cat_path, flower_path)

    expected_shapes = [
        (900, 900, 3),
        (513, 464, 4),
        (100, 100, 4),
        (100, 100, 4),
        (513, 464, 4),
        (900, 900, 3),
    ]

    assert len(images) == 6

    for img, exp_shape in zip(images, expected_shapes):
        arr = img.numpy()
        assert arr.shape == exp_shape
        assert arr.dtype == "uint8"

    assert images.shape == (6, None, None, None)
    assert images.shape_interval.lower == (6, 100, 100, 3)
    assert images.shape_interval.upper == (6, 900, 900, 4)


@enabled_datasets
def test_iterate_compressed_samples(ds: Dataset, cat_path, flower_path):
    images = ds.create_tensor(TENSOR_KEY, htype="image", sample_compression="png")

    assert images.meta.dtype == "uint8"
    assert images.meta.sample_compression == "png"

    _populate_compressed_samples(images, cat_path, flower_path)

    expected_shapes = [
        (900, 900, 3),
        (513, 464, 4),
        (100, 100, 4),
        (100, 100, 4),
        (513, 464, 4),
        (900, 900, 3),
    ]

    assert len(images) == len(expected_shapes)
    for image, expected_shape in zip(images, expected_shapes):
        x = image.numpy()

        assert (
            type(x) == np.ndarray
        ), "Check is necessary in case a `PIL` object is returned instead of an array."
        assert x.shape == expected_shape
        assert x.dtype == "uint8"


@enabled_datasets
def test_uncompressed(ds: Dataset):
    images = ds.create_tensor(TENSOR_KEY, sample_compression=None)

    images.append(np.ones((100, 100, 100)))
    images.extend(np.ones((3, 101, 2, 1)))
    ds.clear_cache()
    np.testing.assert_array_equal(images[0].numpy(), np.ones((100, 100, 100)))
    np.testing.assert_array_equal(images[1:4].numpy(), np.ones((3, 101, 2, 1)))


@pytest.mark.xfail(raises=SampleCompressionError, strict=True)
@pytest.mark.parametrize(
    "bad_shape",
    [
        # raises OSError: cannot write mode LA as JPEG
        (100, 100, 2),
        # raises OSError: cannot write mode RGBA as JPE
        (100, 100, 4),
    ],
)
def test_jpeg_bad_shapes(memory_ds: Dataset, bad_shape):
    # jpeg allowed shapes:
    # ---------------------
    # (100) works!
    # (100,) works!
    # (100, 100) works!
    # (100, 100, 1) works!
    # (100, 100, 2) raises   | OSError: cannot write mode LA as JPEG
    # (100, 100, 3) works!
    # (100, 100, 4) raises   | OSError: cannot write mode RGBA as JPEG
    # (100, 100, 5) raises   | TypeError: Cannot handle this data type: (1, 1, 5), |u1
    # (100, 100, 100) raises | TypeError: Cannot handle this data type: (1, 1, 100), |u1

    tensor = memory_ds.create_tensor(TENSOR_KEY, sample_compression="jpeg")
    tensor.append(np.ones(bad_shape, dtype="uint8"))


def test_compression_aliases(memory_ds: Dataset):
    tensor = memory_ds.create_tensor("jpeg_tensor", sample_compression="jpeg")
    assert tensor.meta.sample_compression == "jpeg"

    tensor = memory_ds.create_tensor("jpg_tensor", sample_compression="jpg")
    assert tensor.meta.sample_compression == "jpeg"


@pytest.mark.xfail(raises=UnsupportedCompressionError, strict=True)
def test_unsupported_compression(memory_ds: Dataset):
    memory_ds.create_tensor(TENSOR_KEY, sample_compression="bad_compression")
    # TODO: same tests but with `dtype`


@pytest.mark.xfail(raises=TensorMetaMissingRequiredValue, strict=True)
def test_missing_sample_compression_for_image(memory_ds: Dataset):
    memory_ds.create_tensor("tensor", htype="image")


@pytest.mark.xfail(raises=TensorMetaMutuallyExclusiveKeysError, strict=True)
def test_sample_chunk_compression_mutually_exclusive(memory_ds: Dataset):
    memory_ds.create_tensor(
        "tensor", htype="image", sample_compression="png", chunk_compression="lz4"
    )


@enabled_datasets
def test_chunkwise_compression(ds: Dataset, cat_path, flower_path):
    images = ds.create_tensor("images", htype="image", chunk_compression="jpg")
    images.append(hub.read(cat_path))
    images.append(hub.read(cat_path))
    expected_arr = np.random.randint(0, 10, (500, 450, 3)).astype("uint8")
    images.append(expected_arr)
    images.append(hub.read(cat_path))
    images.append(hub.read(cat_path))
    expected_img = np.array(hub.read(cat_path))
    assert_images_close(images[2].numpy(), expected_arr)
    for img in images[[0, 1, 3, 4]]:
        assert_images_close(img.numpy(), expected_img)

    images = ds.create_tensor("images2", htype="image", chunk_compression="png")
    images.append(hub.read(flower_path))
    images.append(hub.read(flower_path))
    expected_arr = np.random.randint(0, 256, (500, 450, 4)).astype("uint8")
    images.append(expected_arr)
    images.append(hub.read(flower_path))
    images.append(hub.read(flower_path))
    expected_img = np.array(hub.read(flower_path))

    np.testing.assert_array_equal(images[2].numpy(), expected_arr)
    for img in images[[0, 1, 3, 4]]:
        np.testing.assert_array_equal(img, expected_img)

    labels = ds.create_tensor("labels", chunk_compression="lz4")
    data = [[0] * 50, [1, 2, 3] * 100, [4, 5, 6] * 200, [7, 8, 9] * 300]
    labels.extend(data)
    for row, label in zip(data, labels):
        np.testing.assert_array_equal(row, label.numpy())


@enabled_datasets
@pytest.mark.parametrize("compression", hub.compression.AUDIO_COMPRESSIONS)
def test_audio(ds: Dataset, compression, audio_paths):
    path = audio_paths[compression]
    if path.endswith(".mp3"):
        audio = mp3_read_file_f32(path)
    elif path.endswith(".flac"):
        audio = flac_read_file_f32(path)
    elif path.endswith(".wav"):
        audio = wav_read_file_f32(path)
    arr = np.frombuffer(audio.samples, dtype=np.float32).reshape(
        audio.num_frames, audio.nchannels
    )
    ds.create_tensor("audio", htype="audio", sample_compression=compression)
    with ds:
        for _ in range(10):
            ds.audio.append(hub.read(path))  # type: ignore
    for i in range(10):
        np.testing.assert_array_equal(ds.audio[i].numpy(), arr)  # type: ignore
